#
# Module to compute conditional independencencies in a DAG, and
# sufficient tests.
#
# Author: Jesús Renero
#
import itertools
from typing import List, Optional, Union
import networkx as nx


class ConditionalIndependencies:
    """
    A class to store conditional independencies in a graph.

    Attributes
    ----------
    _cache : dict
        A dictionary representing the conditional independencies.

    Methods
    -------
    add(x, y, z)
        Adds a new conditional independence to the cache.
    __str__()
        Returns a string representation of the conditional independencies.
    __repr__()
        Returns a string representation of the conditional independencies.
    """

    def __init__(self):
        self._cache = {}

    def __str__(self):
        s = "Conditional Independencies:\n"
        for (x, y, z) in self._cache:
            if z == None:
                s += f"  {x} ⊥ {y}\n"
            else:
                s += f"  {x} ⊥ {y} | {z}\n"
        return s

    def __repr__(self) -> str:
        """Returns a string representation of the ConditionalIndependencies object."""
        items = []
        for x, y, z in self._cache.items():
            if z is None:
                items.append(f"({x!r}, {y!r}):{[]!r}")
            else:
                items.append(f"({x!r}, {y!r}):{list(z)!r}")
        return "{" + ", ".join(items) + "}"

    def add(self, var1: str, var2: str,
            conditioning_set: Optional[List[str]] = None) -> None:
        """
        Adds a new conditional independence to the cache.

        Parameters
        ----------
        var1 : str
            A node in the graph.
        var2 : str
            A node in the graph.
        conditioning_set : list of str or None
            A set of nodes in the graph.
        """
        conditioning_set = tuple(conditioning_set) if conditioning_set is not None else None
        if not self._cache:
            self._cache[(var1, var2, conditioning_set)] = True
            return
        if (var1, var2, conditioning_set) in self._cache:
            return
        self._cache[(var1, var2, conditioning_set)] = True


class SufficientSets:
    """
    A class to represent the sufficient sets of a conditional independence test.

    Attributes
    ----------
    _cache : list
        A list of tuples representing the sufficient sets.

    Methods
    -------
    add(suff_set)
        Adds a new sufficient set to the cache.

        Parameters
        ----------
        suff_set : list
            A list of tuples representing the new sufficient set to be added.

    __str__()
        Returns a string representation of the sufficient sets.
    """

    def __init__(self):
        self._cache = []

    def add(self, suff_set):
        """
        Adds a new sufficient set to the cache.

        Parameters
        ----------
        suff_set : list
            A list of tuples representing the new sufficient set to be added.
        """

        for element in suff_set:
            # and ((y, x) not in self._cache):
            if (element not in self._cache):
                self._cache.append(element)

    def __str__(self):
        """
        Returns a string representation of the sufficient sets.

        Returns
        -------
        str
            A string representation of the sufficient sets.
        """
        s = "Sufficient sets:\n"
        if self._cache:
            for sufficient_set in self._cache:
                s += f"  {sufficient_set}\n"
        else:
            s = "No sufficient sets found"

        return s

    def __repr__(self):
        s = "["
        last = False
        if self._cache:
            for sufficient_set in self._cache:
                s += f"{sufficient_set}"
                if sufficient_set == self._cache[-1]:
                    last = True
                if not last:
                    s += ", "
        s += "]"
        return s


def get_backdoor_paths(dag: nx.DiGraph, x: str, y: str):
    """
    Returns all backdoor paths between two nodes in a graph. A backdoor path
    is a path that starts with an edge towards 'x' and ends with an edge
    towards 'y'.

    Parameters:
    -----------
    dag: nx.DiGraph
        A directed graph
    x: str
        A node in the graph
    y: str
        A node in the graph

    Returns:
    --------
    paths: list
        A list of paths between x and y
    """
    # Check if x or y are not in the graph
    if x not in dag.nodes() or y not in dag.nodes():
        return []
    
    # If x and y are the same node, return empty list
    if x == y:
        return []
        
    undirected_graph = dag.to_undirected()
    # list all paths between 'x' and 'y'
    paths = (p for p in nx.all_simple_paths(
        undirected_graph, source=x, target=y) if len(p) > 1 and dag.has_edge(p[1], x))
    return list(paths)


def get_paths(graph: nx.DiGraph, x: str, y: str):
    """
    Returns all simple paths between two nodes in a directed graph.

    Parameters
    ----------
        - graph (nx.DiGraph): A directed graph.
        - x (str): The starting node.
        - y (str): The ending node.

    Returns
    -------
        - list: A list of all simple paths between x and y.
    """
    # Check if x or y are not in the graph
    if x not in graph.nodes() or y not in graph.nodes():
        return []
    
    # If x and y are the same node, return empty list
    if x == y:
        return []
        
    return list(nx.all_simple_paths(graph, source=x, target=y))


def find_colliders_in_path(dag: nx.DiGraph, path: List[str]):
    """
    Returns all colliders in a path.

    Parameters:
    -----------
    G: nx.DiGraph
        A directed graph
    path: list
        A path formed by nodes in the graph

    Returns:
    --------
    colliders: set
        A set of colliders in the path
    """
    colliders = []
    for i in range(1, len(path)-1):
        if dag.has_edge(path[i-1], path[i]) and dag.has_edge(path[i+1], path[i]):
            colliders.append(path[i])

    return set(colliders)


def get_sufficient_sets_for_pair(dag, x, y, verbose=False):
    """
    Compute the sufficient sets for a pair of nodes in a graph. A sufficient set
    is a set of nodes that blocks all backdoor paths between x and y.

    Parameters:
    -----------
    G: nx.DiGraph
        A directed graph
    x: str
        A node in the graph
    y: str
        A node in the graph
    verbose: bool
        If True, print additional information

    Returns:
    --------
    sufficient_sets: list
        A list of sufficient sets for the pair of nodes (x, y)
    """
    backdoor_paths = get_backdoor_paths(dag, x, y)
    if verbose:
        if backdoor_paths:
            print(f" Found {len(backdoor_paths)} backdoor paths")
        else:
            print(" No backdoor paths found")
    sufficient_sets = []
    for path in backdoor_paths:
        print(f"  Checking backdoor path: {path}") if verbose else None
        # get all nodes in the path except the first and last
        sufficient_set = path[1:-1]
        # check that no node in sufficient_set is descendant of x
        descendants = nx.descendants(dag, x)
        if any([d in descendants for d in sufficient_set]):
            if verbose:
                print(
                    f"Path {path} discarded because it contains a descendant of x")
            continue

        sufficient_sets.append(sufficient_set)

    # Now I must check that the nodes in the sufficient set block every backdoor path
    # between x and y
    final_suff_set = []
    for sufficient_set in sufficient_sets:
        if verbose:
            print(
                f"     ",
                f"Checking that {sufficient_set} blocks all backdoor paths "
                f"between {x} and {y}")
        # Check if any of the nodes in the sufficient set is in a collider in the path
        colliders = find_colliders_in_path(dag, [x] + sufficient_set + [y])
        # If any of the nodes in the sufficient set is a collider, then continue
        if len(colliders) > 0:
            if verbose:
                print(
                    f"        {sufficient_set} contains a collider: {colliders}")
            continue
        all_conditions = True
        for path in backdoor_paths:
            if verbose:
                print(f"      Checking path {path}")
            # Check that this path can be blocked by any node in the sufficient set
            # The path is blocked if any of the nodes in the sufficient set
            # is in the path
            colliders = find_colliders_in_path(dag, path)
            if verbose:
                if colliders:
                    print(f"        ! Colliders in path: {colliders}")
                else:
                    print("        - No colliders in path")
            # Find what nodes from the sufficient set are in the path
            nodes_in_path = set(sufficient_set).intersection(set(path))
            if verbose:
                if nodes_in_path:
                    print(
                        f"        + Nodes from sufficient set in path: {nodes_in_path}")
                else:
                    print(f"        - No nodes from sufficient set in path")
            if len(nodes_in_path) > 0:
                # Check that at least one of the nodes in the path is NOT a collider
                if nodes_in_path.intersection(colliders) == set():
                    if verbose:
                        print(
                            f"          Path {path} blocked by nodes in "
                            f"{sufficient_set} ")
                elif len(nodes_in_path) == len(colliders):
                    if verbose:
                        print(f"      ALL nodes in {sufficient_set} are colliders "
                              f"in {path} \n"
                              f"      => {nodes_in_path.intersection(colliders)} == "
                              f"{colliders}")
                    all_conditions = False
                else:
                    if verbose:
                        print(f"      Some nodes in {sufficient_set} are NOT colliders "
                              f"in {path} ")
            else:
                if verbose:
                    print(f"        No nodes in {sufficient_set} are in {path}, "
                          f"so they do not block this path ")
                all_conditions = False
        if all_conditions:
            if verbose:
                print(
                    f"          {sufficient_set} blocks all backdoor paths "
                    f"between x and y")
            final_suff_set.append(sufficient_set)

    return final_suff_set


def get_sufficient_sets(dag, verbose=False):
    """
    Get the sufficient sets (admissible sets) for all pairs of nodes in a graph.

    Parameters:
    -----------
    G: nx.DiGraph
        A directed graph
    verbose: bool
        If True, print additional information

    Returns:
    --------
    suff_sets: Suff_Sets
        A list of sufficient sets for all pairs of nodes in the graph
    """
    suff_sets = SufficientSets()
    for x, y in itertools.combinations(dag.nodes(), 2):
        if verbose:
            print(f"Checking pair ({x}, {y})...", end="", sep="")
        sufficient_set = get_sufficient_sets_for_pair(dag, x, y, verbose)
        if sufficient_set:
            print(f"  Adding sufficient set: {sufficient_set}") if verbose else None
            suff_sets.add(sufficient_set)

    return suff_sets

#
# XXX: This is not used, and should be removed or replaced by the NX 
# implementation of d-separation or DoWhy implementation.
#
def get_conditional_independencies(dag, verbose=False):
    """
    Computes the set of conditional independencies implied by the graph G.

    Parameters:
    -----------
    dag : networkx.DiGraph
        The directed acyclic graph representing the causal relationships
        between the variables.
    verbose : bool, optional
        If True, prints additional information about the computation.

    Returns:
    --------
    cond_indeps : Cond_Indep
        The object containing the set of conditional independencies implied
        by the graph G.
    """
    cond_indeps = ConditionalIndependencies()
    # Enumerate all pairs of nodes in G that are not d_separated
    for x, y in itertools.combinations(dag.nodes(), 2):
        # Check if x and y are connected by an edge
        if dag.has_edge(x, y) or dag.has_edge(y, x):
            continue
        if not nx.d_separated(dag, {x}, {y}, set()):
            if verbose:
                print(f"Pair ({x}, {y})")
            paths = get_paths(dag, x, y)
            # Check if any of the paths contains a collider
            for path in paths:
                if verbose:
                    print("  Path:", path)
                colliders = find_colliders_in_path(dag, path)
                if len(colliders) == 0:
                    blockers = set(path[1:-1])
                    if len(blockers) == 1:
                        cond_indeps.add(x, y, blockers.pop())
                    else:
                        # Check if the set of blockers is exactly the entire graph
                        # without "x" and "y"
                        if blockers != set(dag.nodes()) - {x, y}:
                            cond_indeps.add(x, y, tuple(blockers))
                            if verbose:
                                print(f"    (no colliders on path {path})\n"
                                      f"     {x} ⊥ {y} | {blockers}")
                        else:
                            if verbose:
                                print(f"    The set of blockers is the entire graph. ")
                else:
                    if verbose:
                        print(f"    Colliders in path: {colliders}")
                    for blocker in path[1:-1]:
                        if verbose:
                            print(f"    Blocking on {blocker}")
                        if blocker not in colliders:
                            cond_indeps.add(x, y, blocker)
                            if verbose:
                                print(f"       {x} ⊥ {y} | {blocker}")
                        else:
                            if verbose:
                                print(
                                    f"    Blocking on {blocker} is Collider: "
                                    f"{colliders}")
        else:
            cond_indeps.add(x, y)
            if verbose:
                print(f"Pair ({x}, {y})\n"
                      f"   {x} ⊥ {y} | ∅")

    return cond_indeps


def custom_main():
    G = nx.DiGraph()
    G.add_edges_from(
        [
            ('x1', 'x2'),
            ('x2', 'x3'),
            ('x1', 'x4'),
            ('x2', 'x4')
        ]
    )
    ss = get_sufficient_sets(G, verbose=True)
    print(ss)
    cond_independencies = get_conditional_independencies(G, verbose=True)
    print(cond_independencies)


def main():
    G = nx.DiGraph()
    G.add_edges_from([('z1', 'x'), ('z1', 'z3'), ('z3', 'x'),
                     ('z3', 'y'), ('x', 'y'), ('z2', 'z3'), ('z2', 'y')])

    ss = get_sufficient_sets(G, verbose=True)
    print(ss)

    cond_independencies = get_conditional_independencies(G, verbose=False)
    print(cond_independencies)


def dag_main():
    from causalexplain.estimators.pc.dag import DAG

    G = DAG([('x1', 'x2'), ('x2', 'x3'), ('x1', 'x4'), ('x2', 'x4')])
    ci = G.get_independencies()
    print(ci)


if __name__ == "__main__":
    custom_main()
    # main()
    print("\n-----\n")
    dag_main()



import os
from unittest.mock import MagicMock

import pandas as pd
import pytest

from causalexplain.causalexplainer import GraphDiscovery
from causalexplain.common import DEFAULT_REGRESSORS


@pytest.fixture
def sample_csv(tmp_path):
    # Create a temporary CSV file
    df = pd.DataFrame({
        'A': [1, 2, 3],
        'B': [4, 5, 6],
        'C': [7, 8, 9]
    })
    csv_path = tmp_path / "test_data.csv"
    df.to_csv(csv_path, index=False)
    return str(csv_path)


@pytest.fixture
def sample_dot(tmp_path):
    # Create a temporary DOT file
    dot_content = """
    digraph G {
        A -> B;
        B -> C;
    }
    """
    dot_path = tmp_path / "test_graph.dot"
    with open(dot_path, 'w') as f:
        f.write(dot_content)
    return str(dot_path)


@pytest.fixture
def mock_experiment(monkeypatch):
    """Mock the Experiment class without relying on pytest-mock"""
    mock = MagicMock()
    monkeypatch.setattr('causalexplain.causalexplainer.Experiment', mock)
    return mock


def test_default_initialization():
    """Test initialization with default parameters"""
    gd = GraphDiscovery()
    assert gd.experiment_name is None
    assert gd.estimator == 'rex'
    assert gd.csv_filename is None
    assert gd.dot_filename is None
    assert gd.verbose is False
    assert gd.seed == 42


def test_valid_initialization(sample_csv, sample_dot):
    """Test initialization with valid parameters"""
    gd = GraphDiscovery(
        experiment_name="test_exp",
        model_type="rex",
        csv_filename=sample_csv,
        true_dag_filename=sample_dot,
        verbose=True,
        seed=123
    )
    assert gd.experiment_name == "test_exp"
    assert gd.estimator == "rex"
    assert gd.csv_filename == sample_csv
    assert gd.dot_filename == sample_dot
    assert gd.verbose is True
    assert gd.seed == 123
    assert gd.dataset_name == "test_data"
    assert gd.data_columns == ['A', 'B', 'C']
    assert gd.regressors == DEFAULT_REGRESSORS


def test_initialization_with_non_rex_model(sample_csv, sample_dot):
    """Test initialization with a non-rex model type"""
    gd = GraphDiscovery(
        experiment_name="test_exp",
        model_type="random_forest",
        csv_filename=sample_csv,
        true_dag_filename=sample_dot
    )
    assert gd.regressors == ["random_forest"]


def test_invalid_csv_path():
    """Test initialization with non-existent CSV file"""
    with pytest.raises(FileNotFoundError):
        GraphDiscovery(
            experiment_name="test_exp",
            model_type="rex",
            csv_filename="non_existent.csv",
            true_dag_filename="graph.dot"
        )


def test_partial_initialization():
    """Test initialization with partial parameters"""
    with pytest.raises(ValueError) as exc_info:
        GraphDiscovery(experiment_name="test_exp")
    assert "'experiment_name' and 'csv_filename' must be provided" in str(
        exc_info.value)


def test_mismatched_experiment_name_and_csv():
    """Test that providing only experiment_name or only csv_filename 
    raises ValueError"""
    # Test providing only experiment_name
    with pytest.raises(ValueError) as exc_info:
        GraphDiscovery(experiment_name="test_exp", csv_filename=None)
    assert "'experiment_name' and 'csv_filename' must be provided" in str(
        exc_info.value)

    # Test providing only csv_filename
    with pytest.raises(ValueError) as exc_info:
        GraphDiscovery(experiment_name=None, csv_filename="data.csv")
    assert "'experiment_name' and 'csv_filename' must be provided" in str(
        exc_info.value)


def test_none_values():
    """Test initialization with explicit None values"""
    gd = GraphDiscovery(
        experiment_name=None,
        model_type=None,
        csv_filename=None,
        true_dag_filename=None
    )
    assert gd.estimator == 'rex'
    assert gd.csv_filename is None
    assert gd.dot_filename is None
    assert gd.experiment_name is None


def test_empty_string_values():
    """Test initialization with empty string values"""
    gd = GraphDiscovery(experiment_name="", csv_filename="")
    assert gd.experiment_name is None
    assert gd.csv_filename is None


def test_whitespace_string_values():
    """Test initialization with whitespace string values"""
    gd = GraphDiscovery(experiment_name="   ", csv_filename="  ")
    assert gd.experiment_name is None
    assert gd.csv_filename is None


def test_create_experiments_rex(mock_experiment, sample_csv, sample_dot):
    """Test create_experiments with rex model type"""
    gd = GraphDiscovery(
        experiment_name="test_exp",
        model_type="rex",
        csv_filename=sample_csv,
        true_dag_filename=sample_dot
    )
    trainers = gd.create_experiments()

    # For rex, we should have multiple trainers (one for each DEFAULT_REGRESSOR)
    assert len(trainers) == len(DEFAULT_REGRESSORS)

    # Check that Experiment was called correctly for each regressor
    for model_type in DEFAULT_REGRESSORS:
        trainer_name = f"test_data_{model_type}"
        assert trainer_name in trainers
        mock_experiment.assert_any_call(
            experiment_name="test_data",
            csv_filename=sample_csv,
            dot_filename=sample_dot,
            model_type=model_type,
            input_path=os.path.dirname(sample_csv),
            output_path=os.getcwd(),
            verbose=False
        )


def test_create_experiments_non_rex(mock_experiment, sample_csv, sample_dot):
    """Test create_experiments with non-rex model type"""
    gd = GraphDiscovery(
        experiment_name="test_exp",
        model_type="lingam",
        csv_filename=sample_csv,
        true_dag_filename=sample_dot
    )
    trainers = gd.create_experiments()

    # For non-rex, we should have only one trainer
    assert len(trainers) == 1
    trainer_name = "test_data_lingam"
    assert trainer_name in trainers

    # Check that Experiment was called correctly
    mock_experiment.assert_called_once_with(
        experiment_name="test_data",
        csv_filename=sample_csv,
        dot_filename=sample_dot,
        model_type="lingam",
        input_path=os.path.dirname(sample_csv),
        output_path=os.getcwd(),
        verbose=False
    )


def test_create_experiments_without_initialization():
    """Test create_experiments without proper initialization"""
    gd = GraphDiscovery()
    with pytest.raises(AttributeError):
        gd.create_experiments()


def test_create_experiments_empty_regressors(mock_experiment, sample_csv, sample_dot):
    """Test create_experiments with empty regressors list"""
    gd = GraphDiscovery(
        experiment_name="test_exp",
        model_type="rex",
        csv_filename=sample_csv,
        true_dag_filename=sample_dot
    )
    gd.regressors = []  # Empty the regressors list
    trainers = gd.create_experiments()

    assert len(trainers) == 0
    mock_experiment.assert_not_called()



"""
REX (Regression with Explainability) estimator module.

This module provides the REX estimator, which uses regression models and 
explainability techniques to discover causal relationships in data.
"""

from .rex import Rex
from .knowledge import Knowledge

__all__ = [
    'Rex',
    'Knowledge'
]



from _pytest.tmpdir import tmp_path
import networkx as nx
import pandas as pd
import numpy as np
import pytest
import torch
import os
import pickle
import math

from causalexplain.common import utils


class TestSelectDevice:
    # Tests that the function returns 'cuda' when force is 'cuda' and cuda is
    # available
    def test_returns_cuda_when_force_is_cuda_and_cuda_is_not_available(self):
        # Try to force cuda when cuda is not available
        # Should raise a ValueError
        with pytest.raises(ValueError):
            utils.select_device(force='cuda')

    # Tests that the function returns 'mps' when force is 'mps' and mps is
    # available

    def test_returns_mps_when_force_is_mps_and_mps_is_available(self):
        result = utils.select_device(force='mps')
        if torch.backends.mps.is_available():
            assert result == 'mps'
        else:
            assert result == 'cpu'

    # Tests that the function returns 'cpu' when force is 'cpu' and no other
    # options are available

    def test_returns_cpu_when_force_is_cpu_and_no_other_options_are_available(self):
        result = utils.select_device(force='cpu')
        assert result == 'cpu'

    # Tests that the function returns 'mps' when mps is available and no force
    # is specified

    def test_returns_mps_when_mps_is_available_and_no_force_is_specified(self):
        result = utils.select_device()
        if torch.backends.mps.is_available():
            assert result == 'mps'
        else:
            assert result == 'cpu'

    # Tests that the function returns 'cpu' when no other options are available
    # and no force is specified

    def test_returns_cpu_when_no_other_options_are_available_and_no_force_is_specified(self):
        result = utils.select_device()
        if torch.backends.mps.is_available():
            assert result == 'mps'
        else:
            assert result == 'cpu'


class TestGraphIntersection:

    # Test with two identical graphs

    def test_identical_graphs(self):
        # Create two identical graphs
        g1 = nx.DiGraph()
        g1.add_node(1)
        g1.add_node(2)
        g1.add_edge(1, 2)

        g2 = nx.DiGraph()
        g2.add_node(1)
        g2.add_node(2)
        g2.add_edge(1, 2)

        # Call the graph_intersection function
        result = utils.graph_intersection(g1, g2)

        # Check if the result is equal to the input graphs
        assert result.nodes == g1.nodes
        assert result.edges == g1.edges

    # Test with two graphs with different nodes and edges
    def test_different_nodes_edges(self):
        # Create two graphs with different nodes and edges
        g1 = nx.DiGraph()
        g1.add_node(1)
        g1.add_node(2)
        g1.add_edge(1, 2)

        g2 = nx.DiGraph()
        g2.add_node(3)
        g2.add_node(4)
        g2.add_edge(3, 4)

        # Call the graph_intersection function
        result = utils.graph_intersection(g1, g2)

        # Check if the result is an empty graph
        assert len(result.nodes) == 0
        assert len(result.edges) == 0

    # Test with two graphs with some common nodes and edges
    def test_common_nodes_edges(self):
        # Create two graphs with some common nodes and edges
        g1 = nx.DiGraph()
        g1.add_node(1)
        g1.add_node(2)
        g1.add_edge(1, 2)

        g2 = nx.DiGraph()
        g2.add_node(2)
        g2.add_node(3)
        g2.add_edge(2, 3)

        # Call the graph_intersection function
        result = utils.graph_intersection(g1, g2)

        # Check if the result has the common nodes and edges
        assert len(result.nodes) == 1
        assert len(result.edges) == 0
        assert 2 in result.nodes

    # Test two graphs with different node attributes

    def test_different_node_attributes(self):
        # Create two graphs with different node attributes
        g1 = nx.DiGraph()
        g1.add_node(1, attr='A')
        g1.add_node(2, attr='B')
        g1.add_edge(1, 2)

        g2 = nx.DiGraph()
        g2.add_node(1, attr='C')
        g2.add_node(2, attr='D')
        g2.add_edge(1, 2)

        # Call the graph_intersection function
        result = utils.graph_intersection(g1, g2)

        # Check if the result has the minimum node attributes
        assert len(result.nodes) == 2
        # assert 'A' in result.nodes[1]['attr']
        # assert 'B' in result.nodes[2]['attr']

    # Test with two empty graphs

    def test_empty_graphs(self):
        # Create two empty graphs
        g1 = nx.DiGraph()
        g2 = nx.DiGraph()

        # Call the graph_intersection function
        result = utils.graph_intersection(g1, g2)

        # Check if the result is an empty graph
        assert len(result.nodes) == 0
        assert len(result.edges) == 0

    # Test with two undirected graphs

    def test_identical_undirected_graphs(self):
        # Create two identical undirected graphs
        g1 = nx.Graph()
        g1.add_node(1)
        g1.add_node(2)
        g1.add_edge(1, 2)

        g2 = nx.Graph()
        g2.add_node(1)
        g2.add_node(2)
        g2.add_edge(1, 2)

        # Call the graph_intersection function
        result = utils.graph_intersection(g1, g2)

        # Check if the result is equal to the input graphs
        assert result.nodes == g1.nodes
        assert result.edges == g1.edges


class TestGraphUnion:
    """Test suite for graph_union function"""

    def test_union_disjoint_graphs(self):
        g1 = nx.DiGraph()
        g1.add_edges_from([("A", "B"), ("B", "C")])
        g2 = nx.DiGraph()
        g2.add_edges_from([("D", "E"), ("E", "F")])

        union_graph = utils.graph_union(g1, g2)

        assert isinstance(union_graph, nx.DiGraph)
        assert set(union_graph.nodes()) == {"A", "B", "C", "D", "E", "F"}
        assert set(union_graph.edges()) == {("A", "B"), ("B", "C"), ("D", "E"),
                                            ("E", "F")}

    def test_union_overlapping_graphs(self):
        g1 = nx.DiGraph()
        g1.add_edges_from([("A", "B"), ("B", "C")])
        nx.set_node_attributes(g1, {"A": {"weight": 1}, "B": {
                               "weight": 2}, "C": {"weight": 3}})

        g2 = nx.DiGraph()
        g2.add_edges_from([("B", "C"), ("C", "D")])
        nx.set_node_attributes(g2, {"B": {"weight": 3}, "C": {
                               "weight": 4}, "D": {"weight": 5}})

        union_graph = utils.graph_union(g1, g2)

        assert isinstance(union_graph, nx.DiGraph)
        assert set(union_graph.nodes()) == {"A", "B", "C", "D"}
        assert set(union_graph.edges()) == {("A", "B"), ("B", "C"), ("C", "D")}
        assert union_graph.nodes["B"]["weight"] == 2.5
        assert union_graph.nodes["C"]["weight"] == 3.5

    def test_union_with_empty_graph(self):
        g1 = nx.DiGraph()
        g1.add_edges_from([("A", "B")])
        g2 = nx.DiGraph()  # Empty graph

        union_graph = utils.graph_union(g1, g2)

        assert isinstance(union_graph, nx.DiGraph)
        assert set(union_graph.nodes()) == {"A", "B"}
        assert set(union_graph.edges()) == {("A", "B")}


class TestClassifyVariable:
    """Test suite for _classify_variable function"""

    def test_binary_numeric(self):
        """Test binary classification with numeric values"""
        arr = pd.Series([0, 1, 0, 1, 0])
        assert utils._classify_variable(arr) == "binary"

        arr = pd.Series([1.0, 2.0, 1.0, 2.0])
        assert utils._classify_variable(arr) == "binary"

    def test_binary_categorical(self):
        """Test binary classification with categorical values"""
        arr = pd.Series(['yes', 'no', 'yes', 'no'])
        assert utils._classify_variable(arr) == "multiclass"

        arr = pd.Series(['true', 'false', 'true'])
        assert utils._classify_variable(arr) == "multiclass"

    def test_multiclass_categorical(self):
        """Test multiclass classification with categorical values"""
        arr = pd.Series(['a', 'b', 'c', 'd'])
        assert utils._classify_variable(arr) == "multiclass"

        arr = pd.Series(['low', 'medium', 'high', 'medium', 'low'])
        assert utils._classify_variable(arr) == "multiclass"

    def test_continuous_numeric(self):
        """Test continuous classification with numeric values"""
        arr = pd.Series([1.0, 2.5, 3.7, 4.2, 5.0])
        assert utils._classify_variable(arr) == "continuous"

        arr = pd.Series(np.random.normal(0, 1, 100))
        assert utils._classify_variable(arr) == "continuous"

    def test_mixed_types(self):
        """Test with mixed types (should be handled as multiclass)"""
        arr = pd.Series(['1', '2', 'a', 'b'])
        assert utils._classify_variable(arr) == "multiclass"

    def test_special_cases(self):
        """Test special cases"""
        # Empty series
        arr = pd.Series([])
        assert utils._classify_variable(arr) == "continuous"

        # Series with NaN values
        arr = pd.Series([1, 2, np.nan, 4])
        assert utils._classify_variable(arr) == "continuous"

        # Boolean values
        arr = pd.Series([True, False, True])
        assert utils._classify_variable(arr) == "binary"


class TestCastCategoricalsToInt:
    """Test suite for cast_categoricals_to_int function"""

    def test_cast_multiclass(self):
        """Test casting multiclass categorical variables"""
        df = pd.DataFrame({
            'A': pd.Categorical(['a', 'b', 'c', 'a']),
            'B': pd.Categorical(['x', 'y', 'z', 'x']),
            'C': [1.0, 2.0, 3.0, 1.0]  # Numeric column should remain unchanged
        })
        result = utils.cast_categoricals_to_int(df)
        # Categories should be converted to integers starting from 0
        assert result['A'].dtype == np.int16
        assert result['B'].dtype == np.int16
        assert result['C'].dtype == np.float64  # Should remain unchanged
        assert set(result['A'].unique()) == {0, 1, 2}
        assert set(result['B'].unique()) == {0, 1, 2}

    def test_cast_binary(self):
        """Test casting binary categorical variables"""
        df = pd.DataFrame({
            'A': ['yes', 'no', 'yes', 'no'],
            'B': ['true', 'false', 'true', 'false'],
            'C': [0, 1, 0, 1]  # Already binary numeric
        })
        result = utils.cast_categoricals_to_int(df)
        assert result['A'].dtype == np.int16
        assert result['B'].dtype == np.int16
        assert result['C'].dtype == np.int16  # Should remain numeric
        assert set(result['A'].unique()) == {0, 1}
        assert set(result['B'].unique()) == {0, 1}
        assert set(result['C'].unique()) == {0, 1}

    def test_no_categorical(self):
        """Test with no categorical columns"""
        df = pd.DataFrame({
            'A': [1.0, 2.0, 3.0],
            'B': [4.0, 5.0, 6.0],
            'C': [7, 8, 9]
        })
        result = utils.cast_categoricals_to_int(df)
        pd.testing.assert_frame_equal(result, df)  # Should be unchanged

    def test_mixed_types(self):
        """Test with mixed types in the same column"""
        df = pd.DataFrame({
            'A': ['1', '2', 'a', 'b'],  # Mixed strings
            'B': [1, 2, 'three', 'four'],  # Mixed numbers and strings
            'C': [1.0, 2.0, 3.0, 4.0]  # Pure numeric
        })
        result = utils.cast_categoricals_to_int(df)
        assert result['A'].dtype == np.int16  # Should be converted
        assert result['B'].dtype == np.int16  # Should be converted
        assert result['C'].dtype == np.float64  # Should remain unchanged
        assert len(result['A'].unique()) == 4
        assert len(result['B'].unique()) == 4


class TestSaveExperiment:
    """Test suite for save_experiment function"""
    tmp_path = "/tmp"

    def test_save_experiment_creates_folder(self):
        folder = os.path.join(self.tmp_path, "new_folder")
        obj_name = "test_experiment"
        results = {"key0": "value0"}

        output = utils.save_experiment(obj_name, str(folder), results)

        assert os.path.exists(folder)
        assert os.path.exists(output)

    def test_save_experiment_overwrite(self):
        folder = os.path.join(self.tmp_path, "new_folder")
        obj_name = "test_experiment"
        results = {"key2": "valu2e"}

        # Create an existing file
        existing_file = os.path.join(folder, f"{obj_name}.pickle")
        with open(existing_file, 'wb') as f:
            pickle.dump({}, f)

        output = utils.save_experiment(
            obj_name, str(folder), results, overwrite=True)

        assert os.path.exists(output)
        with open(output, 'rb') as f:
            data = pickle.load(f)
        assert data == results

    def test_save_experiment_no_overwrite(self):
        folder = os.path.join(self.tmp_path, "new_folder")
        obj_name = "test_experiment"
        results = {"key3": "value3"}

        # Create an existing file
        existing_file = os.path.join(folder, f"{obj_name}.pickle")
        with open(existing_file, 'wb') as f:
            pickle.dump({}, f)

        output = utils.save_experiment(
            obj_name, str(folder), results, overwrite=False)

        assert os.path.exists(output)
        with open(output, 'rb') as f:
            data = pickle.load(f)
        assert data == results

    def test_save_experiment_return_value(self):
        folder = os.path.join(self.tmp_path, "new_folder")
        obj_name = "test_experiment"
        results = {"key": "value"}

        output = utils.save_experiment(
            obj_name, str(folder), results, overwrite=True)

        expected_output = os.path.join(folder, f"{obj_name}.pickle")
        assert output == str(expected_output)

    def test_save_experiment_saves_correct_data(self):
        folder = os.path.join(self.tmp_path, "new_folder")
        obj_name = "test_experiment"
        results = {"key": "value"}

        output = utils.save_experiment(
            obj_name, str(folder), results, overwrite=True)

        with open(output, 'rb') as f:
            data = pickle.load(f)
        assert data == results

    def test_save_experiment_with_pickle_extension(self):
        folder = os.path.join(self.tmp_path, "pickle_folder")
        obj_name = "test_experiment.pickle"  # Already has .pickle extension
        results = {"key": "value"}

        output = utils.save_experiment(
            obj_name, str(folder), results, overwrite=True)

        # Should not add another .pickle
        expected_output = os.path.join(folder, obj_name)
        assert output == str(expected_output)
        assert os.path.exists(output)
        with open(output, 'rb') as f:
            data = pickle.load(f)
        assert data == results


class TestLoadExperiment:
    """Test suite for load_experiment function"""
    tmp_path = "/tmp"

    def test_load_experiment_existing_file(self):
        folder = os.path.join(self.tmp_path, "existing_folder")
        obj_name = "test_experiment"
        expected_results = {"key1": "value1"}

        # Ensure the folder exists
        os.makedirs(folder, exist_ok=True)

        # Create a pickle file with expected results
        with open(os.path.join(folder, f"{obj_name}.pickle"), 'wb') as f:
            pickle.dump(expected_results, f)

        # Load the experiment
        loaded_obj = utils.load_experiment(obj_name, folder)

        assert loaded_obj == expected_results

    def test_load_experiment_non_existent_file(self):
        folder = os.path.join(self.tmp_path, "non_existent_folder")
        obj_name = "non_existent_experiment"

        with pytest.raises(FileNotFoundError):
            utils.load_experiment(obj_name, folder)


class TestValidOutputName:
    """Test suite for valid_output_name function"""
    tmp_path = "/tmp"

    def test_valid_output_name_no_conflict(self):
        path = os.path.join(self.tmp_path, "unique_folder")
        filename = "unique_file"
        extension = "txt"

        # Ensure the folder exists
        os.makedirs(path, exist_ok=True)

        # Test for a filename with no conflict
        result = utils.valid_output_name(filename, path, extension)
        expected = os.path.join(path, f"{filename}.{extension}")

        assert result == expected

    def test_valid_output_name_with_conflict(self):
        path = os.path.join(self.tmp_path, "conflict_folder")
        filename = "conflict_file"
        extension = "txt"

        # Ensure the folder exists
        os.makedirs(path, exist_ok=True)

        # Create a conflicting file
        with open(os.path.join(path, f"{filename}.{extension}"), 'w') as f:
            f.write("dummy content")

        # Test for a filename with conflict
        result = utils.valid_output_name(filename, path, extension)
        expected = os.path.join(path, f"{filename}_1.{extension}")

        assert result == expected

    def test_valid_output_name_no_extension(self):
        path = os.path.join(self.tmp_path, "no_ext_folder")
        filename = "no_ext_file"

        # Ensure the folder exists
        os.makedirs(path, exist_ok=True)

        # Test for a filename without extension
        result = utils.valid_output_name(filename, path)
        expected = os.path.join(path, filename)

        assert result == expected

    def test_valid_output_name_with_pickle_extension(self):
        path = os.path.join(self.tmp_path, "pickle_ext_folder")
        filename = "test_file.pickle"  # Already has .pickle extension
        extension = "pickle"

        # Ensure the folder exists
        os.makedirs(path, exist_ok=True)

        # Test for a filename that already has the extension
        result = utils.valid_output_name(filename, path, extension)
        # Should not add another .pickle
        expected = os.path.join(path, filename)

        assert result == expected


class TestGraphFromDotFile:
    """Test suite for graph_from_dot_file function"""
    tmp_path = "/tmp"

    def test_graph_from_valid_dot_file(self):
        dot_content = """
        strict digraph G {
            concentrate = true;
            
            A -> B;
            B -> C;
            C -> A;
        }
        """
        dot_file_path = os.path.join(self.tmp_path, "valid_graph.dot")

        with open(dot_file_path, 'w') as f:
            f.write(dot_content)

        graph = utils.graph_from_dot_file(dot_file_path)

        assert graph is not None
        assert isinstance(graph, nx.DiGraph)
        assert set(graph.nodes()) == {'A', 'B', 'C'}
        assert set(graph.edges()) == {('A', 'B'), ('B', 'C'), ('C', 'A')}

    def test_graph_from_non_existent_file(self):
        non_existent_path = os.path.join(self.tmp_path, "non_existent.dot")
        graph = utils.graph_from_dot_file(non_existent_path)
        assert graph is None

    def test_graph_from_invalid_dot_file(self):
        invalid_dot_content = """
        digraph G {
            A -> B
            B ->
        }
        """
        dot_file_path = os.path.join(self.tmp_path, "invalid_graph.dot")

        with open(dot_file_path, 'w') as f:
            f.write(invalid_dot_content)

        graph = utils.graph_from_dot_file(dot_file_path)
        assert graph is None


class TestGraphFromDictionary:
    """Test suite for graph_from_dictionary function"""

    def test_graph_from_simple_dictionary(self):
        simple_dict = {
            'u': ['v', 'w'],
            'x': ['y']
        }
        graph = utils.graph_from_dictionary(simple_dict)

        assert isinstance(graph, nx.DiGraph)
        assert set(graph.nodes()) == {'u', 'v', 'w', 'x', 'y'}
        assert set(graph.edges()) == {('v', 'u'), ('w', 'u'), ('y', 'x')}

    def test_graph_from_weighted_dictionary(self):
        weighted_dict = {
            'u': [('v', 0.2), ('w', 0.7)],
            'x': [('y', 0.5)]
        }
        graph = utils.graph_from_dictionary(weighted_dict)

        assert isinstance(graph, nx.DiGraph)
        assert set(graph.nodes()) == {'u', 'v', 'w', 'x', 'y'}
        assert set(graph.edges()) == {('v', 'u'), ('w', 'u'), ('y', 'x')}
        assert graph['v']['u']['weight'] == 0.2
        assert graph['w']['u']['weight'] == 0.7
        assert graph['y']['x']['weight'] == 0.5

    def test_graph_from_empty_dictionary(self):
        empty_dict = {}
        graph = utils.graph_from_dictionary(empty_dict)

        assert isinstance(graph, nx.DiGraph)
        assert len(graph.nodes()) == 0
        assert len(graph.edges()) == 0


class TestGraphFromAdjacency:
    """Test suite for graph_from_adjacency function"""

    def test_graph_from_default_adjacency(self):
        adjacency = np.array([
            [0, 1, 0],
            [0, 0, 1],
            [1, 0, 0]
        ])
        graph = utils.graph_from_adjacency(adjacency)

        assert isinstance(graph, nx.DiGraph)
        assert set(graph.nodes()) == {0, 1, 2}
        assert set(graph.edges()) == {(0, 1), (1, 2), (2, 0)}

    def test_graph_with_threshold(self):
        adjacency = np.array([
            [0, 0.5, 0],
            [0, 0, 0.2],
            [0.3, 0, 0]
        ])
        graph = utils.graph_from_adjacency(adjacency, th=0.25)

        assert isinstance(graph, nx.DiGraph)
        assert set(graph.nodes()) == {0, 1, 2}
        assert set(graph.edges()) == {(0, 1), (2, 0)}

    def test_graph_with_inverse_edges(self):
        adjacency = np.array([
            [0, 1, 0],
            [0, 0, 1],
            [1, 0, 0]
        ])
        graph = utils.graph_from_adjacency(adjacency, inverse=True)

        assert isinstance(graph, nx.DiGraph)
        assert set(graph.nodes()) == {0, 1, 2}
        assert set(graph.edges()) == {(1, 0), (2, 1), (0, 2)}

    def test_graph_with_absolute_values(self):
        adjacency = np.array([
            [0, -0.5, 0],
            [0, 0, -0.2],
            [0.3, 0, 0]
        ])
        graph = utils.graph_from_adjacency(
            adjacency, absolute_values=True, th=0.25)

        assert isinstance(graph, nx.DiGraph)
        assert set(graph.nodes()) == {0, 1, 2}
        assert set(graph.edges()) == {(0, 1), (2, 0)}

    def test_graph_with_node_labels(self):
        adjacency = np.array([
            [0, 1, 0],
            [0, 0, 1],
            [1, 0, 0]
        ])
        labels = ['A', 'B', 'C']
        graph = utils.graph_from_adjacency(adjacency, node_labels=labels)

        assert isinstance(graph, nx.DiGraph)
        assert set(graph.nodes()) == {'A', 'B', 'C'}
        assert set(graph.edges()) == {('A', 'B'), ('B', 'C'), ('C', 'A')}


class TestGraphFromAdjacencyFile:
    """Test suite for graph_from_adjacency_file function"""
    tmp_path = "/tmp"

    def test_graph_from_file_with_header(self):
        file_path = os.path.join(self.tmp_path, "adjacency_with_header.csv")
        content = """A,B,C\n0,1,0\n0,0,1\n1,0,0"""

        with open(file_path, 'w') as f:
            f.write(content)

        graph, df = utils.graph_from_adjacency_file(file_path)

        assert isinstance(graph, nx.DiGraph)
        assert set(graph.nodes()) == {'A', 'B', 'C'}
        assert set(graph.edges()) == {('A', 'B'), ('B', 'C'), ('C', 'A')}
        assert isinstance(df, pd.DataFrame)

    def test_graph_from_file_without_header(self):
        file_path = os.path.join(self.tmp_path, "adjacency_no_header.csv")
        content = """0,1,0\n0,0,1\n1,0,0"""

        with open(file_path, 'w') as f:
            f.write(content)

        labels = ['A', 'B', 'C']
        graph, df = utils.graph_from_adjacency_file(
            file_path, labels=labels, header=False)

        assert isinstance(graph, nx.DiGraph)
        assert set(graph.nodes()) == {'A', 'B', 'C'}
        assert set(graph.edges()) == {('A', 'B'), ('B', 'C'), ('C', 'A')}
        assert isinstance(df, pd.DataFrame)

    def test_graph_with_threshold_from_file(self):
        file_path = os.path.join(self.tmp_path, "adjacency_threshold.csv")
        content = """A,B,C\n0,0.5,0\n0,0,0.2\n0.3,0,0"""

        with open(file_path, 'w') as f:
            f.write(content)

        graph, df = utils.graph_from_adjacency_file(file_path, th=0.25)

        assert isinstance(graph, nx.DiGraph)
        assert set(graph.nodes()) == {'A', 'B', 'C'}
        assert set(graph.edges()) == {('A', 'B'), ('C', 'A')}
        assert isinstance(df, pd.DataFrame)

    def test_graph_with_custom_labels_from_file(self):
        file_path = os.path.join(self.tmp_path, "adjacency_custom_labels.csv")
        content = """0,1,0\n0,0,1\n1,0,0"""

        with open(file_path, 'w') as f:
            f.write(content)

        custom_labels = ['X', 'Y', 'Z']
        graph, df = utils.graph_from_adjacency_file(
            file_path, labels=custom_labels, header=False)

        assert isinstance(graph, nx.DiGraph)
        assert set(graph.nodes()) == {'X', 'Y', 'Z'}
        assert set(graph.edges()) == {('X', 'Y'), ('Y', 'Z'), ('Z', 'X')}
        assert isinstance(df, pd.DataFrame)


class TestGraphToAdjacency:
    """Test suite for graph_to_adjacency function"""

    def test_simple_graph_to_adjacency(self):
        graph = nx.DiGraph()
        graph.add_edge('A', 'B', weight=1)
        graph.add_edge('B', 'C', weight=1)
        graph.add_edge('C', 'A', weight=1)
        labels = ['A', 'B', 'C']

        adjacency_matrix = utils.graph_to_adjacency(graph, labels)

        expected_matrix = np.array([
            [0, 1, 0],
            [0, 0, 1],
            [1, 0, 0]
        ])

        assert np.array_equal(adjacency_matrix, expected_matrix)

    def test_graph_with_custom_weight_label(self):
        graph = nx.DiGraph()
        graph.add_edge('A', 'B', custom_weight=2)
        graph.add_edge('B', 'C', custom_weight=3)
        graph.add_edge('C', 'A', custom_weight=4)
        labels = ['A', 'B', 'C']

        adjacency_matrix = utils.graph_to_adjacency(
            graph, labels, weight_label='custom_weight')

        expected_matrix = np.array([
            [0, 2, 0],
            [0, 0, 3],
            [4, 0, 0]
        ])

        assert np.array_equal(adjacency_matrix, expected_matrix)

    def test_graph_to_adjacency_with_missing_labels(self):
        graph = nx.DiGraph()
        graph.add_edge('A', 'B', weight=1)
        graph.add_edge('B', 'C', weight=1)
        labels = ['A', 'B', 'C', 'D']  # 'D' is a missing node in the graph

        adjacency_matrix = utils.graph_to_adjacency(graph, labels)

        expected_matrix = np.array([
            [0, 1, 0, 0],
            [0, 0, 1, 0],
            [0, 0, 0, 0],
            [0, 0, 0, 0]
        ])

        assert np.array_equal(adjacency_matrix, expected_matrix)


class TestGraphToAdjacencyFile:
    """Test suite for graph_to_adjacency_file function"""
    tmp_path = "/tmp"

    def test_write_simple_graph_to_file(self):
        graph = nx.DiGraph()
        graph.add_edge('A', 'B', weight=1)
        graph.add_edge('B', 'C', weight=1)
        graph.add_edge('C', 'A', weight=1)
        labels = ['A', 'B', 'C']
        file_path = os.path.join(self.tmp_path, "simple_graph.csv")

        utils.graph_to_adjacency_file(graph, file_path, labels)

        with open(file_path, 'r') as f:
            content = f.read()

        expected_content = """A,B,C
A,0.0,1.0,0.0
B,0.0,0.0,1.0
C,1.0,0.0,0.0
"""
        assert content == expected_content

    def test_write_graph_with_custom_weight_label_to_file(self):
        graph = nx.DiGraph()
        graph.add_edge('A', 'B', weight=2)
        graph.add_edge('B', 'C', weight=3)
        graph.add_edge('C', 'A', weight=4)
        labels = ['A', 'B', 'C']
        file_path = os.path.join(self.tmp_path, "custom_weight_graph.csv")

        utils.graph_to_adjacency_file(graph, file_path, labels)

        with open(file_path, 'r') as f:
            content = f.read()

        expected_content = """A,B,C
A,0.0,2.0,0.0
B,0.0,0.0,3.0
C,4.0,0.0,0.0
"""
        assert content == expected_content

    def test_write_graph_with_missing_labels_to_file(self):
        graph = nx.DiGraph()
        graph.add_edge('A', 'B', weight=1)
        graph.add_edge('B', 'C', weight=1)
        labels = ['A', 'B', 'C', 'D']  # 'D' is a missing node in the graph
        file_path = os.path.join(self.tmp_path, "missing_labels_graph.csv")

        utils.graph_to_adjacency_file(graph, file_path, labels)

        with open(file_path, 'r') as f:
            content = f.read()

        expected_content = """A,B,C,D
A,0.0,1.0,0.0,0.0
B,0.0,0.0,1.0,0.0
C,0.0,0.0,0.0,0.0
D,0.0,0.0,0.0,0.0
"""
        assert content == expected_content


class TestCorrectEdgeFromPrior:
    """Test suite for correct_edge_from_prior function"""

    def test_remove_edge_in_top_list(self):
        dag = nx.DiGraph()
        dag.add_edge('A', 'B')
        prior = [['A', 'B'], ['C']]

        orientation = utils.correct_edge_from_prior(
            dag, 'A', 'B', prior, verbose=False)

        assert orientation == +1
        assert not dag.has_edge('A', 'B')

    def test_add_edge_in_correct_direction(self):
        dag = nx.DiGraph()
        prior = [['A'], ['B', 'C']]

        orientation = utils.correct_edge_from_prior(
            dag, 'A', 'B', prior, verbose=False)

        assert orientation == +1
        assert dag.has_edge('A', 'B')

    def test_reverse_edge(self):
        dag = nx.DiGraph()
        prior = [['B'], ['A']]

        orientation = utils.correct_edge_from_prior(
            dag, 'A', 'B', prior, verbose=False)

        assert orientation == -1
        assert not dag.has_edge('B', 'A')

    def test_no_change_for_unclear_order(self):
        dag = nx.DiGraph()
        dag.add_edge('A', 'B')
        prior = [['A'], ['C'], ['B']]

        orientation = utils.correct_edge_from_prior(
            dag, 'A', 'B', prior, verbose=False)

        assert orientation == 1
        assert dag.has_edge('A', 'B')

    def test_nodes_not_in_prior(self):
        dag = nx.DiGraph()
        dag.add_edge('X', 'Y')
        prior = [['A'], ['B']]

        orientation = utils.correct_edge_from_prior(
            dag, 'X', 'Y', prior, verbose=False)

        assert orientation == 0
        assert dag.has_edge('X', 'Y')

    def test_edge_reflects_backward_connection(self):
        dag = nx.DiGraph()
        dag.add_edge('A', 'B')
        prior = []

        orientation = utils.correct_edge_from_prior(
            dag, 'A', 'B', prior, verbose=False)

        assert orientation == 0

    def test_edge_reflects_connection_in_same_layer(self):
        dag = nx.DiGraph()
        dag.add_edge('B', 'C')
        prior = [['A'], ['B', 'C']]

        orientation = utils.correct_edge_from_prior(
            dag, 'B', 'C', prior, verbose=False)

        assert orientation == 0


class TestValidCandidatesFromPrior:

    def test_valid_candidates_with_prior(self):
        feature_names = ['A', 'B', 'C', 'D']
        effect = 'C'
        prior = [['A'], ['B', 'C'], ['D']]

        candidates = utils.valid_candidates_from_prior(
            feature_names, effect, prior)

        assert candidates == ['A', 'B']

    def test_effect_not_in_prior(self):
        feature_names = ['A', 'B', 'C', 'D']
        effect = 'E'
        prior = [['A'], ['B', 'C'], ['D']]

        with pytest.raises(ValueError, match="Effect 'E' not found in prior"):
            utils.valid_candidates_from_prior(feature_names, effect, prior)

    def test_valid_candidates_without_prior(self):
        feature_names = ['A', 'B', 'C', 'D']
        effect = 'C'
        prior = None

        candidates = utils.valid_candidates_from_prior(
            feature_names, effect, prior)

        assert candidates == ['A', 'B', 'D']

    def test_empty_prior(self):
        feature_names = ['A', 'B', 'C', 'D']
        effect = 'C'
        prior = []

        candidates = utils.valid_candidates_from_prior(
            feature_names, effect, prior)

        assert candidates == ['A', 'B', 'D']


class TestBreakCyclesUsingPrior:
    def test_break_cycles_removes_edges(self):
        # Create a graph with a cycle
        dag = nx.DiGraph()
        dag.add_edges_from([('A', 'B'), ('B', 'C'), ('C', 'A')])
        prior = [['A'], ['B'], ['C']]

        # Break cycles using prior
        new_dag = utils.break_cycles_using_prior(dag, prior)

        # Assert that the cycle is broken
        assert not list(nx.simple_cycles(new_dag)
                        ), "The cycle should be broken."

    def test_break_cycles_keeps_valid_edges(self):
        # Create a graph without a cycle
        dag = nx.DiGraph()
        dag.add_edges_from([('A', 'B'), ('B', 'C')])
        prior = [['A'], ['B'], ['C']]

        # Break cycles using prior
        new_dag = utils.break_cycles_using_prior(dag, prior)

        # Assert that no edges are removed
        assert list(new_dag.edges) == [
            ('A', 'B'), ('B', 'C')], "Edges should remain unchanged."

    def test_break_cycles_with_no_prior(self):
        # Create a graph with a cycle
        dag = nx.DiGraph()
        dag.add_edges_from([('A', 'B'), ('B', 'C'), ('C', 'A')])
        prior = None

        # Break cycles using prior
        new_dag = utils.break_cycles_using_prior(dag, prior)

        # Assert that the cycle is broken
        assert sorted(list(nx.simple_cycles(new_dag))[0]) == ['A', 'B', 'C'], \
            "The cycle should remain."


# Mock class for discrepancies
class MockDiscrepancy:
    def __init__(self, shap_gof):
        self.shap_gof = shap_gof


class TestPotentialMisorientedEdges:

    def test_misoriented_edges_identification(self):
        # Define a loop and discrepancies
        loop = ['A', 'B', 'C']
        discrepancies = {
            'A': {'B': MockDiscrepancy(0.3), 'C': MockDiscrepancy(0.8)},
            'B': {'A': MockDiscrepancy(0.2), 'C': MockDiscrepancy(0.1)},
            'C': {'A': MockDiscrepancy(0.6), 'B': MockDiscrepancy(0.9)}
        }

        # Identify misoriented edges
        misoriented_edges = utils.potential_misoriented_edges(
            loop, discrepancies)

        # Assert the correct misoriented edges are identified
        expected_edges = [('A', 'B', 0.1)]
        assert misoriented_edges[0][0] == expected_edges[0][0]
        assert misoriented_edges[0][1] == expected_edges[0][1]
        assert math.isclose(
            misoriented_edges[0][2], expected_edges[0][2], rel_tol=1e-6)

    def test_no_misoriented_edges(self):
        # Define a loop and discrepancies
        loop = ['A', 'B', 'C']
        discrepancies = {
            'A': {'B': MockDiscrepancy(0.2), 'C': MockDiscrepancy(0.8)},
            'B': {'A': MockDiscrepancy(0.3), 'C': MockDiscrepancy(0.1)},
            'C': {'A': MockDiscrepancy(0.6), 'B': MockDiscrepancy(0.9)}
        }

        # Identify misoriented edges
        misoriented_edges = utils.potential_misoriented_edges(
            loop, discrepancies)

        # Assert no misoriented edges are identified
        assert misoriented_edges == []


class TestBreakCyclesIfPresent:

    def test_no_cycles(self):
        dag = nx.DiGraph()
        dag.add_edges_from([("A", "B"), ("B", "C")])
        discrepancies = {}
        result = utils.break_cycles_if_present(dag, discrepancies)
        assert list(result.edges) == [
            ("A", "B"), ("B", "C")], "DAG should remain unchanged"

    def test_single_cycle(self):
        dag = nx.DiGraph()
        dag.add_edges_from([("A", "B"), ("B", "C"), ("C", "A")])
        discrepancies = {
            "B": {"A": MockDiscrepancy(0.2)},
            "C": {"B": MockDiscrepancy(0.1)},
            "A": {"C": MockDiscrepancy(0.3)}
        }
        result = utils.break_cycles_if_present(dag, discrepancies)
        assert not list(nx.simple_cycles(result)), "Cycles should be broken"

    def test_multiple_cycles(self):
        dag = nx.DiGraph()
        dag.add_edges_from(
            [("A", "B"), ("B", "C"), ("C", "A"), ("C", "D"), ("D", "B")])
        discrepancies = {
            "B": {"A": MockDiscrepancy(0.2), "D": MockDiscrepancy(0.5)},
            "C": {"B": MockDiscrepancy(0.1)},
            "A": {"C": MockDiscrepancy(0.3)},
            "D": {"C": MockDiscrepancy(0.4)}
        }
        result = utils.break_cycles_if_present(dag, discrepancies)
        assert not list(nx.simple_cycles(result)
                        ), "All cycles should be broken"

    def test_with_prior_knowledge(self):
        dag = nx.DiGraph()
        dag.add_edges_from([("A", "B"), ("B", "C"), ("C", "A")])
        discrepancies = {
            "B": {"A": MockDiscrepancy(0.2)},
            "C": {"B": MockDiscrepancy(0.1)},
            "A": {"C": MockDiscrepancy(0.3)}
        }
        prior = [["A"], ["B"], ["C"]]
        result = utils.break_cycles_if_present(dag, discrepancies, prior=prior)
        assert not list(nx.simple_cycles(result)
                        ), "Cycles should be broken using prior knowledge"

    def test_potential_misoriented_edges(self):
        dag = nx.DiGraph()
        dag.add_edges_from([("A", "B"), ("B", "C"), ("C", "A")])
        discrepancies = {
            "B": {"A": MockDiscrepancy(0.2)},
            "C": {"B": MockDiscrepancy(0.1)},
            "A": {"C": MockDiscrepancy(0.3)}
        }
        result = utils.break_cycles_if_present(dag, discrepancies)
        assert not list(nx.simple_cycles(result)
                        ), "Cycles should be broken by changing orientation"


class TestGraphToDotFile:

    tmp_path = "/tmp"

    def test_graph_to_dot_file(self):
        # Create a simple directed graph
        graph = nx.DiGraph()
        graph.add_edges_from([(1, 2), (2, 3)])

        # Define the output file path
        output_file = os.path.join(self.tmp_path, "test_graph.dot")

        # Use the function to write the graph to a dot file
        utils.graph_to_dot_file(graph, output_file)

        # Read the file back and verify its contents
        with open(output_file, 'r') as f:
            content = f.read()
            assert 'digraph' in content
            assert '1 -> 2' in content
            assert '2 -> 3' in content



"""
A module to run experiments with the causalexplain package, and simplify the
process of loading and saving experiments in notebooks.

Example:
    >> from causalexplain.common.notebook import Experiment
    >> experiment = Experiment("linear", csv_filename="linear.csv")
    >> rex = experiment.load()

(C) 2023, 2024 J. Renero
"""

import os
import time
import warnings
from os import path

import numpy as np
import pandas as pd
from sklearn.preprocessing import StandardScaler

from . import utils
from ..estimators.cam.cam import CAM
from ..estimators.fci.fci import FCI
from ..estimators.ges.ges import GES
from ..estimators.lingam.lingam import DirectLiNGAM as LiNGAM
from ..estimators.pc.pc import PC
from ..estimators.notears.notears import NOTEARS
from ..estimators.rex.rex import Rex

warnings.filterwarnings('ignore')


# pylint: disable=E1101:no-member, W0201:attribute-defined-outside-init, W0511:fixme
# pylint: disable=C0103:invalid-name
# pylint: disable=C0116:missing-function-docstring
# pylint: disable=R0913:too-many-arguments
# pylint: disable=R0914:too-many-locals, R0915:too-many-statements
# pylint: disable=W0106:expression-not-assigned, R1702:too-many-branches


estimators = {
    'rex': Rex,
    'fci': FCI,
    'pc': PC,
    'lingam': LiNGAM,
    'ges': GES,
    'cam': CAM,
    'notears': NOTEARS
}
method_names = ['pc', 'fci', 'ges', 'lingam', 'cam', 'notears', 'rex']
metric_columns = ['method', 'data_type', 'f1', 'precision',
                  'recall', 'aupr', 'Tp', 'Tn', 'Fp', 'Fn', 'shd', 'sid',
                  'n_edges', 'ref_n_edges', 'diff_edges', 'name']
RAW_DAG_NAMES = ['G_shap', 'G_prior', 'G_iter', 'G_iter_prior']
COMBINED_DAG_NAMES = ['un_G_shap', 'in_G_shap',
                      'un_G_prior', 'in_G_prior',
                      'un_G_iter', 'in_G_iter',
                      'un_G_iter_prior', 'in_G_iter_prior']


class BaseExperiment:
    """
    Base class for experiments.

    Args:
    input_path (str): The path to the input data.
    output_path (str): The path to save the experiment output.
    train_anyway (bool, optional): Whether to train the model even if the
        experiment exists. Defaults to False.
    save_anyway (bool, optional): Whether to save the experiment even
        if it exists. Defaults to False.
    train_size (float, optional): The proportion of data to use for training.
        Defaults to 0.9.
    random_state (int, optional): The random state for reproducibility.
        Defaults to 42.
    verbose (bool, optional): Whether to display verbose output.
        Defaults to False.
    """

    def __init__(
            self,
            input_path: str,
            output_path: str,
            train_anyway: bool = False,
            save_anyway: bool = False,
            scale: bool = False,
            train_size: float = 0.9,
            random_state: int = 42,
            verbose: bool = False):

        self.input_path = input_path
        self.output_path = output_path
        self.train_anyway = train_anyway
        self.save_anyway = save_anyway
        self.scale = scale
        self.train_size = train_size
        self.random_state = random_state
        self.verbose = verbose

        # Display Options
        np.set_printoptions(precision=4, linewidth=100)
        pd.set_option('display.precision', 4)
        pd.set_option('display.float_format', '{:.4f}'.format)

    def prepare_experiment_input(
            self,
            experiment_filename,
            csv_filename=None,
            dot_filename=None):
        """
        - Loads the data and
        - splits it into train and test,
        - scales it
        - loads the reference graph from the dot file, which has to be named
          as the experiment file, with the .dot extension
        """
        self.experiment_name = path.splitext(
            path.basename(experiment_filename))[0]
        if csv_filename is None:
            csv_filename = f"{path.join(self.input_path, self.experiment_name)}.csv"
        if dot_filename is None:
            dot_filename = f"{path.join(self.input_path, self.experiment_name)}.dot"

        self.data = pd.read_csv(csv_filename)
        self.data = self.data.apply(pd.to_numeric, downcast='float')
        if self.scale:
            scaler = StandardScaler()
            self.data = pd.DataFrame(
                scaler.fit_transform(self.data), columns=self.data.columns)
            self.train_data = self.data.sample(
                frac=self.train_size, random_state=self.random_state)
            self.test_data = self.data.drop(self.train_data.index)
        else:
            self.train_data = self.data.sample(
                frac=self.train_size, random_state=self.random_state)
            self.test_data = self.data.drop(self.train_data.index)

        self.ref_graph = utils.graph_from_dot_file(dot_filename)

    def experiment_exists(self, name):
        """Checks whether the experiment exists in the output path"""
        return os.path.exists(
            os.path.join(self.output_path, f"{os.path.basename(name)}.pickle"))

    def create_estimator(self, estimator_name: str, name: str, **kwargs):
        """
        Dynamically creates an instance of a class based on the estimator name.

        Args:
        estimator_name (str): The name of the estimator (key in the 'estimators'
            dictionary).
        name (str): The name of the estimator instance.
        *args: Variable length argument list to be passed to the class constructor.
        **kwargs: Arbitrary keyword arguments to be passed to the class constructor.

        Returns:
        An instance of the specified class, or None if the class does not exist.
        """
        estimator_class = estimators.get(estimator_name)
        if estimator_class is None:
            print(f"Estimator '{estimator_name}' not found.")
            return None

        # Special case: when estimator is ReX, model_type needs also to be
        # passed to the constructor
        if estimator_name == 'rex':
            kwargs['model_type'] = self.model_type

        return estimator_class(name=name, **kwargs)


class Experiment(BaseExperiment):
    """
    Represents an experiment for causal graph analysis.

    Methods:
        load: Loads the experiment data.
        fit: Fits the experiment data.
        save: Saves the experiment data.
    """

    estimator_name = None

    def __init__(
        self,
        experiment_name,
        csv_filename: str = None,
        dot_filename: str = None,
        model_type: str = 'nn',
        input_path="/Users/renero/phd/data/RC4/",
        output_path="/Users/renero/phd/output/RC4/",
        train_size: float = 0.9,
        random_state: int = 42,
        verbose=False
    ):
        """
        Initializes a new instance of the Experiment class.

        Args:
            experiment_name (str): The name of the experiment.
            csv_filename (str, optional): The filename of the CSV file containing
                the data. Defaults to None.
            dot_filename (str, optional): The filename of the DOT file containing
                the causal graph. Defaults to None.
            model_type (str, optional): The type of model to use. Defaults to 'nn'.
                Other options are: 'gbt', 'nn', 'cam', 'pc', 'fci', 'notears',
                'ges' and 'lingam'.
            input_path (str, optional): The path to the input data.
                Defaults to "/Users/renero/phd/data/RC4/".
            output_path (str, optional): The path to save the output.
                Defaults to "/Users/renero/phd/output/RC4/".
            train_size (float, optional): The proportion of data to use for training.
                Defaults to 0.9.
            random_state (int, optional): The random seed for reproducibility.
                Defaults to 42.
            verbose (bool, optional): Whether to print verbose output.
                Defaults to False.
        """

        super().__init__(
            input_path, output_path, train_size=train_size,
            random_state=random_state, verbose=verbose)
        self.model_type = self._check_model_type(model_type)
        self.is_fitted = False
        self.verbose = verbose

        # Prepare the input
        self.prepare_experiment_input(
            experiment_name, csv_filename, dot_filename)

    def _check_model_type(self, model_type):
        """
        Checks if the model type is valid.
        """
        model_type = model_type.lower()
        if model_type in ['dnn', 'nn']:
            model_type = 'nn'
        elif model_type == 'gbt':
            model_type = 'gbt'
        elif model_type in method_names:
            model_type = model_type
        else:
            raise ValueError(
                f"Model type '{model_type}' not supported. "
                f"Supported options are: "
                f"'nn', 'gbt', 'pc', 'fci', 'cam', 'notears', 'ges' and "
                f"'lingam'.")

        return model_type

    def fit(self, estimator_name='rex', **kwargs):
        """
        Fits the experiment data.

        Args:
            **kwargs: Additional keyword arguments to pass to the Rex constructor.

        Returns:
            Rex: The fitted experiment data.
        """
        self.estimator_name = estimator_name
        kwargs['model_type'] = self.model_type

        estimator_object = self.create_estimator(
            estimator_name, name=self.experiment_name, **kwargs)

        pipeline = kwargs.pop('pipeline') if 'pipeline' in kwargs else None

        estimator_object.fit(
            self.train_data, y=self.test_data, pipeline=pipeline)

        setattr(self, estimator_name, estimator_object)
        self.is_fitted = True

        return self

    def predict(self, estimator='rex', **kwargs):
        """
        Predicts with the experiment data.

        Args:
            **kwargs: Additional keyword arguments to pass to the `predict()` method

        Returns:
            Rex: The fitted experiment data.
        """
        estimator = getattr(self, self.estimator_name)
        estimator.predict(self.train_data, **kwargs)

        return self

    def fit_predict(self, estimator='rex', **kwargs):
        """
        Fits and predicts with the experiment data.

        Args:
            **kwargs: Additional keyword arguments to pass to the Rex constructor.

        Returns:
            Rex: The fitted experiment data.
        """
        start_time = time.time()
        self.estimator_name = estimator

        # Extract from the kwargs the parameter 'prior'
        prior = kwargs.pop('prior') if 'prior' in kwargs else None

        estimator_object = self.create_estimator(
            estimator, name=self.experiment_name, **kwargs)
        fit_predict_args = [self.train_data, self.test_data, self.ref_graph]
        if estimator == 'rex':
            fit_predict_args.append(prior)
        if estimator_object is not None:
            estimator_object.fit_predict(*fit_predict_args)
        else:
            raise ValueError(f"Estimator '{estimator}' not found.")

        setattr(self, estimator, estimator_object)
        end_time = time.time()
        self.fit_predict_time = end_time - start_time

        return self

    def load(self, exp_name=None) -> "Experiment":
        """
        Loads the experiment data.

        Args:
            exp_name (str, optional): The name of the experiment to load.
            If None, loads the current experiment. Defaults to None.

        Returns:
            Rex: The loaded experiment data.
        """

        if exp_name is None:
            exp_name = self.experiment_name

        if self.model_type:
            exp_object = utils.load_experiment(
                f"{exp_name}_{self.model_type}", self.output_path)
        else:
            exp_object = utils.load_experiment(exp_name, self.output_path)

        # A priori, I don't know which estimator was used to train the experiment
        # so I have to check the type of the object
        if isinstance(exp_object, Rex):
            self.estimator_name = 'rex'
        elif isinstance(exp_object, PC):
            self.estimator_name = 'pc'
        elif isinstance(exp_object, LiNGAM):
            self.estimator_name = 'lingam'
        elif isinstance(exp_object, GES):
            self.estimator_name = 'ges'
        elif isinstance(exp_object, FCI):
            self.estimator_name = 'fci'
        elif isinstance(exp_object, CAM):
            self.estimator_name = 'cam'
        elif isinstance(exp_object, NOTEARS):
            self.estimator_name = 'notears'
        else:
            raise ValueError(
                f"Estimator '{exp_name}' not recognized.")

        setattr(self, self.estimator_name, exp_object)
        setattr(self, 'estimator', exp_object)

        if self.verbose:
            print(f"Loaded '{exp_name}' ({self.model_type.upper()}) "
                  f"from '{self.output_path}'")
            fit_time = utils.format_time(self.rex.fit_time)
            predict_time = utils.format_time(self.rex.predict_time)
            print(f"This model took {fit_time[0]:.1f}{fit_time[1]}. to fit, and "
                  f"{predict_time[0]:.1f}{predict_time[1]}. to build predicted DAGs")

        return self

    def save(self, exp_name=None, overwrite: bool = False):
        """
        Saves the experiment data.

        Args:
        -----
        - exp_name (str, optional): The name to save the experiment as.
            If None, uses the experiment name. Defaults to None.
        - overwrite (bool, optional): Whether to overwrite an existing
            experiment with the same name. Defaults to False.
        """
        save_as = exp_name if exp_name is not None else self.experiment_name
        where_to = utils.save_experiment(
            f"{save_as}_{self.model_type}",
            self.output_path, getattr(self, self.estimator_name),
            overwrite)

        if self.verbose:
            print(f"Saved '{self.experiment_name}' to '{where_to}'")

        return where_to


if __name__ == "__main__":
    np.set_printoptions(precision=4, linewidth=150)
    warnings.filterwarnings('ignore')
    extra_args = {
        'rex': {
            'prog_bar': True,
            'verbose': False,
            'hpo_n_trials': 1,
            'bootstrap_trials': 10,
            'bootstrap_parallel_jobs': -1,
            'parallel_jobs': -1
        },
        'pc': {},
        'ges': {},
        'lingam': {},
        'fci': {},
        'cam': {
            'pruning': True,
            'pruneMethodPars': {"cutOffPVal": 0.05, "numBasisFcts": 10}
        },
        'notears': {}
    }

    input_path = os.path.expanduser("~/phd/data/")
    output_path = os.path.expanduser("~/phd/output/")

    method_name = "rex"
    dataset_name = "toy_dataset"
    # dataset_name =  "generated_10vars_linear_0"

    exp = Experiment(
        experiment_name=dataset_name,
        csv_filename=os.path.join(input_path, f"{dataset_name}.csv"),
        dot_filename=os.path.join(input_path, f"{dataset_name}.dot"),
        model_type="gbt",
        input_path=input_path,
        output_path=output_path)

    exp = exp.fit_predict(method_name, **extra_args[method_name])
    method = getattr(exp, method_name)
    print(method.dag.edges())
    print(method.metrics)
    t, u = utils.format_time(exp.fit_predict_time)
    print(f"Elapsed time: {t:.1f}{u}")



import types

import numpy as np
import pytest
import torch

from causalexplain.estimators.notears import lbfgsb_scipy


def test_init_requires_single_param_group():
    p1 = torch.nn.Parameter(torch.tensor([1.0]))
    p2 = torch.nn.Parameter(torch.tensor([2.0]))
    with pytest.raises(ValueError):
        lbfgsb_scipy.LBFGSBScipy([{"params": [p1]}, {"params": [p2]}])


def test_step_invokes_closure_and_applies_solution(monkeypatch):
    linear = torch.nn.Linear(1, 1, bias=False)
    with torch.no_grad():
        linear.weight[:] = torch.tensor([[0.5]])

    called = {}

    def fake_minimize(fun, x0, method=None, jac=None, bounds=None):
        # emulate scipy returning an updated flat parameter vector
        loss, grad = fun(x0)
        called["loss"] = loss
        called["grad"] = grad
        result = types.SimpleNamespace()
        result.x = x0 + 0.25  # pretend optimizer nudged parameters
        return result

    monkeypatch.setattr(lbfgsb_scipy.sopt, "minimize", fake_minimize)

    optimizer = lbfgsb_scipy.LBFGSBScipy(linear.parameters())

    def closure():
        optimizer.zero_grad()
        output = linear(torch.tensor([[2.0]]))
        loss = (output ** 2).sum()
        loss.backward()
        return loss

    optimizer.step(closure)

    assert pytest.approx(called["loss"]) == 1.0
    assert called["grad"].shape == (1,)
    assert np.allclose(linear.weight.detach().numpy(), np.array([[0.75]]))



from typing import List, Set, Union
import numpy as np
from scipy import stats
from sklearn.base import BaseEstimator


class RegQuality(BaseEstimator):

    def __init__(self):
        super().__init__()

    @staticmethod
    def predict(
            scores: List[float],
            gamma_shape: float = 1,
            gamma_scale: float = 1,
            threshold: float = 0.9,
            verbose: bool = False) -> Set[int]:
        """
        Returns the indices of features that are both gamma and outliers. Both criteria
        are applied to the given scores to determine if the MSE error obtained from
        the regression is bad compared with the rest of regressions for the other features
        in the dataset, and thus the feature should be considered a parent node.

        Parameters
        ----------
        scores: List[float]
            List of scores

        Returns
        -------
        Set[int]
            List of indices of features that are both gamma and outliers
        """
        scores = np.array(scores)
        gamma_indices = RegQuality._gamma_criteria(
            scores, gamma_shape, gamma_scale, threshold, verbose=verbose)
        outliers_indices = RegQuality._mad_criteria(scores, verbose=verbose)

        return set(gamma_indices).intersection(outliers_indices)

    @staticmethod
    def _mad_criteria(scores, verbose=False) -> Set[int]:
        """
        Returns indices of outliers in the given scores, using the MAD method.
        Taken from https://stats.stackexchange.com/a/78617
        
        Parameters
        ----------
        scores: List[float]
            List of scores
        verbose: bool
            Whether to print the score and M for each score
            
        Returns
        -------
        Set[int]
            List of indices of outliers in the given scores
        """
        median = np.median(scores)
        ad = [np.abs(score-median) for score in scores]
        mad = np.median(ad)
        M = [np.abs(.6745 * (score-median) / mad) for score in scores]
        mad_indices = [idx for idx, m in enumerate(M) if m > 3.5]

        if verbose:
            print(f"Median: {median:.4f}")
            print(f"Median of absolute differences: {mad:.4f}")
            for score, m in zip(scores, M):
                print(f"Score: {score:.4f}, M: {m:.4f}")

        return set(mad_indices)

    @staticmethod
    def _gamma_criteria(
            scores,
            gamma_shape=1,
            gamma_scale=1,
            threshold=0.9,
            verbose=False) -> Union[None, Set[int]]:
        """
        Returns a list of booleans indicating whether the score is below the threshold

        Parameters
        ----------
        scores: List[float]
            List of scores
        gamma_shape: float
            Shape parameter for gamma distribution
        gamma_scale: float
            Scale parameter for gamma distribution
        threshold: float
            Threshold for gamma criteria
        verbose: bool
            Whether to print the score and gamma criteria for each score

        Returns
        -------
        Set[int]
            List of booleans indicating whether the score is below the threshold
        """
        gamma_indices = []
        for idx, score in enumerate(scores):
            pdf = stats.gamma.pdf(score, a=gamma_shape, scale=gamma_scale)
            if pdf <= threshold:
                gamma_indices.append(idx)
            if verbose:
                print(
                    f"Score: {score:.4f}, PDF: {pdf:.4f}, criteria: {pdf < threshold}")

        return set(gamma_indices)



"""
This file includes all the plot methods for the causal graph

(C) J. Renero, 2022, 2023
"""

# pylint: disable=E1101:no-member, W0201:attribute-defined-outside-init, W0511:fixme
# pylint: disable=C0103:invalid-name, E1121:too-many-function-args
# pylint: disable=C0116:missing-function-docstring, W0212:protected-access
# pylint: disable=R0913:too-many-arguments, disable:W0212:protected-access
# pylint: disable=R0914:too-many-locals, R0915:too-many-statements
# pylint: disable=W0106:expression-not-assigned, R1702:too-many-branches

from copy import copy
from typing import Any, Callable, List, Tuple

import matplotlib as mpl
import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import pandas as pd
import pydotplus
import seaborn as sns
from matplotlib import axes
from matplotlib.backends.backend_pdf import PdfPages
from matplotlib.colors import ListedColormap
from matplotlib.ticker import MultipleLocator
from pydot import Dot
from scipy.cluster.hierarchy import dendrogram
from sklearn.base import BaseEstimator
from sklearn.preprocessing import StandardScaler

from causalexplain.metrics.compare_graphs import evaluate_graph
from .metrics_config import global_nc_metric_types, global_metric_types

metric_labels = {
    'mlp': 'DFN',
    'gbt': 'GBT',
    'intersection': r'$\textrm{DAG}_\cap$',
    'union': r'$\textrm{DAG}_\cup$',
    'union_all': 'all',
    'int_indep': '∩i',
    'int_final': '∩f',
    'union_indep': '∪i',
    'union_final': '∪f',
    'mlp_nc': 'DFN',
    'gbt_nc': 'GBT',
    'intersection_nc': '∩',
    'union_nc': '∪',
    'union_all_nc': 'all'
}
score_titles = {
    'f1': r'$\textrm{F1}$',
    'precision': r'$\textrm{Precision}$',
    'recall': r'$\textrm{Recall}$',
    'aupr': r'$\textrm{AuPR}$',
    'Tp': r'$\textrm{TP}$',
    'Tn': r'$\textrm{TN}$',
    'Fp': r'$\textrm{FP}$',
    'Fn': r'$\textrm{FN}$',
    'shd': r'$\textrm{SHD}$',
    'sid': r'$\textrm{SID}$',
    'n_edges': r'$\textrm{Nr. Edges}$',
    'ref_n_edges': r'$\textrm{Edges in Ground Truth}$',
    'diff_edges': r'$\textrm{Diff. Edges}$',
}
method_labels = {
    'nn': r'$\textrm{R\textsc{e}X}_{\textrm{\tiny DFN}}$',
    'rex_mlp': r'$\textrm{R\textsc{e}X}_{\textrm{\tiny DFN}}$',
    'nn_adj': r'$\textrm{R\textsc{e}X}_{\textrm{\tiny DFN}}^{\textrm{\tiny adj}}$',
    'rex_mlp_adj': r'$\textrm{R\textsc{e}X}_{\textrm{\tiny DFN}}^{\textrm{\tiny adj}}$',
    'gbt': r'$\textrm{R\textsc{e}X}_{\textrm{\tiny GBT}}$',
    'rex_gbt': r'$\textrm{R\textsc{e}X}_{\textrm{\tiny GBT}}$',
    'gbt_adj': r'$\textrm{R\textsc{e}X}_{\textrm{\tiny GBT}}^{\textrm{\tiny adj}}$',
    'rex_gbt_adj': r'$\textrm{R\textsc{e}X}_{\textrm{\tiny GBT}}^{\textrm{\tiny adj}}$',
    'union': r'$\textrm{R\textsc{e}X}_{\cup}$',
    'rex_union': r'$\textrm{R\textsc{e}X}_{\cup}$',
    'union_adj': r'$\textrm{R\textsc{e}X}_{\cup}^{\textrm{\tiny adj}}$',
    'rex_union_adj': r'$\textrm{R\textsc{e}X}_{\cup}^{\textrm{\tiny adj}}$',
    'rex_union_adjnc': r'$\textrm{R\textsc{e}X}_{\cup}^{\textrm{\tiny adj}}$',
    'intersection': r'$\textrm{R\textsc{e}X}_{\cap}$',
    'rex_intersection': r'$\textrm{R\textsc{e}X}_{\cap}$',
    'intersection_adj': r'$\textrm{R\textsc{e}X}_{\cap}^{\textrm{\tiny adj}}$',
    'rex_intersection_adj': r'$\textrm{R\textsc{e}X}_{\cap}^{\textrm{\tiny adj}}$',
    'rex_intersection_adjnc': r'$\textrm{R\textsc{e}X}_{\cap}^{\textrm{\tiny adj}}$',
    'pc': r'$\textrm{PC}$',
    'fci': r'$\textrm{FCI}$',
    'ges': r'$\textrm{GES}$',
    'lingam': r'$\textrm{LiNGAM}$',
    'cam': r'$\textrm{CAM}$',
    'notears': r'$\textrm{NOTEARS}$',
    'G_pc': r'$\textrm{PC}$',
    'G_fci': r'$\textrm{FCI}$',
    'G_ges': r'$\textrm{GES}$',
    'G_lingam': r'$\textrm{LiNGAM}$',
    'G_cam': r'$\textrm{CAM}$',
    'G_notears': r'$\textrm{NOTEARS}$',
    'un_G_iter': r'$\textrm{R\textsc{e}X}$'
}
synth_data_types = ['linear', 'polynomial', 'gp_add', 'gp_mix', 'sigmoid_add']
synth_data_labels = ['Linear', 'Polynomial',
                     'Gaussian(add)', 'Gaussian(mix)', 'Sigmoid(add)']


# Defaults for the graphs plotted
formatting_kwargs = {
    "node_size": 800,
    "node_color": "white",
    "alpha": 0.8,
    "edgecolors": "black",
    "font_weight": "bold",
    "font_family": "monospace",
    "horizontalalignment": "center",
    "verticalalignment": "center_baseline",
    "with_labels": True
}


def setup_plot(**kwargs):  # tex=True, font="serif", dpi=75, font_size=10):
    """Customize figure settings.

    Args:
        tex (bool, optional): use LaTeX. Defaults to True.
        font (str, optional): font type. Defaults to "serif".
        dpi (int, optional): dots per inch. Defaults to 180.
    """
    font_size = kwargs.pop("font_size", 10)
    usetex = kwargs.pop("usetex", True)
    font_familiy = kwargs.pop("font_family", "serif")
    dpi = kwargs.pop("dpi", 75)
    file_format = kwargs.pop("file_format", "pdf")
    title_size = kwargs.pop("title_size", 8)
    axis_labelsize = kwargs.pop("axis_labelsize", 8)
    xtick_labelsize = kwargs.pop("xtick_labelsize", 8)
    ytick_labelsize = kwargs.pop("ytick_labelsize", 8)
    legend_fontsize = kwargs.pop("legend_fontsize", 8)
    axes_labelpad = kwargs.pop("axes_labelpad", 4)
    xtick_majorpad = kwargs.pop("xtick_majorpad", 3)
    ytick_majorpad = kwargs.pop("ytick_majorpad", 3)

    plt.rcParams.update(
        {
            "font.size": font_size,
            "font.family": font_familiy,
            "text.usetex": usetex,
            "figure.subplot.top": 0.9,
            "figure.subplot.right": 0.9,
            "figure.subplot.left": 0.15,
            "figure.subplot.bottom": 0.12,
            "figure.subplot.hspace": 0.4,
            "figure.dpi": dpi,
            "savefig.dpi": 180,
            "savefig.format": file_format,
            "axes.titlesize": title_size,
            "axes.labelsize": axis_labelsize,
            "axes.axisbelow": True,
            "axes.labelpad": axes_labelpad,
            "xtick.direction": "in",
            "ytick.direction": "in",
            "xtick.major.size": 5,
            "xtick.minor.size": 2.25,
            "xtick.major.pad": xtick_majorpad,
            "xtick.minor.pad": 7.5,
            "ytick.major.pad": ytick_majorpad,
            "ytick.minor.pad": 7.5,
            "ytick.major.size": 5,
            "ytick.minor.size": 2.25,
            "xtick.labelsize": xtick_labelsize,
            "ytick.labelsize": ytick_labelsize,
            "legend.fontsize": legend_fontsize,
            "legend.framealpha": 1,
            "figure.titlesize": 12,
            "lines.linewidth": 2,
        }
    )


def add_grid(ax, lines=True, locations=None):
    """Add a grid to the current plot.

    Args:
        ax (Axis): axis object in which to draw the grid.
        lines (bool, optional): add lines to the grid. Defaults to True.
        locations (tuple, optional):
            (xminor, xmajor, yminor, ymajor). Defaults to None.
    """

    if lines:
        ax.grid(lines, alpha=0.5, which="minor", ls=":")
        ax.grid(lines, alpha=0.7, which="major")

    if locations is not None:

        assert (
            len(locations) == 4
        ), "Invalid entry for the locations of the markers"

        xmin, xmaj, ymin, ymaj = locations

        ax.xaxis.set_minor_locator(MultipleLocator(xmin))
        ax.xaxis.set_major_locator(MultipleLocator(xmaj))
        ax.yaxis.set_minor_locator(MultipleLocator(ymin))
        ax.yaxis.set_major_locator(MultipleLocator(ymaj))


def subplots(
        plot_func: Callable,
        *plot_args: Any,
        **kwargs: Any) -> None:
    """
    Plots a set of subplots.

    Arguments:
    ----------
        plot_func: function
            The function to be used to plot each subplot.
        *plot_args: List
            The arguments to be passed to the plot function.
        **kwargs: Dict
            Additional arguments to be passed to the plot function.

    Returns:
    --------
        fig: Figure
            The figure containing the subplots.
    """
    figsize = kwargs.pop("figsize", (8, 6))
    title = kwargs.pop("title", None)
    num_cols = kwargs.pop("num_cols", 4)
    pdf_filename = kwargs.pop("pdf_filename", None)
    setup_plot(**kwargs)
    num_rows = len(plot_args) // num_cols
    if len(plot_args) % num_cols != 0:
        num_rows += 1

    def blank(ax):
        """
        Create a blank subplot.
        """
        npArray = np.array([[[255, 255, 255, 255]]], dtype="uint8")
        ax.imshow(npArray, interpolation="nearest")
        ax.set_axis_off()

    def ax_index(i, j):
        """
        Return the axis index, considering special cases where the number of rows
        or columns is 1.
        """
        nonlocal ax
        if num_rows == 1:
            return ax[j]
        if num_cols == 1:
            return ax[i]
        return ax[i][j]

    plt.rcParams.update({'font.size': 8})
    fig, ax = plt.subplots(num_rows, num_cols, figsize=figsize)
    for i in range(num_rows):
        for j in range(num_cols):
            index = i * num_cols + j
            if index < len(plot_args):
                # axe = ax[i][j]
                plot_func(plot_args[index], ax=ax_index(i, j))
            else:
                blank(ax_index(i, j))

    if title is not None:
        fig.suptitle(title)

    if pdf_filename is not None:
        plt.tight_layout(rect=[0, 0.0, 1, 0.97])
        plt.savefig(pdf_filename, bbox_inches='tight')
        plt.close()
    else:
        plt.tight_layout()
        plt.show()


def format_graph(
    G: nx.DiGraph,
    Gt: nx.DiGraph = None,
    ok_color="green",
    inv_color="lightgreen",
    wrong_color="black",
    missing_color=None
) -> nx.DiGraph:
    if Gt is None:
        for u, v in G.edges():
            G[u][v]['color'] = "black"
            G[u][v]['width'] = 1.0
            G[u][v]['style'] = 'solid'
            G[u][v]['alpha'] = 0.7
    else:
        for u, v in G.edges():
            if Gt.has_edge(u, v):
                G[u][v]['color'] = ok_color
                G[u][v]['width'] = 3.0
                G[u][v]['style'] = 'solid'
                G[u][v]['alpha'] = 1.0
            elif Gt.has_edge(v, u):           # The edge exists, but reversed
                G[u][v]['color'] = inv_color
                G[u][v]['width'] = 2.0
                G[u][v]['style'] = 'dashed'
                G[u][v]['alpha'] = 0.8
            else:                             # The edge does not exist
                G[u][v]['color'] = wrong_color
                G[u][v]['width'] = 1.0
                G[u][v]['style'] = 'dashdot'
                G[u][v]['alpha'] = 0.6
        if missing_color is not None:
            for u, v in Gt.edges():
                if not G.has_edge(u, v) and not G.has_edge(v, u):
                    G.add_edge(u, v)
                    G[u][v]['color'] = missing_color
                    G[u][v]['width'] = 1.0
                    G[u][v]['style'] = 'dotted'
                    G[u][v]['alpha'] = 0.5
    return G


def draw_graph_subplot(
        G: nx.DiGraph,
        root_causes: list = None,
        layout: dict = None,
        title: str = None,
        ax: plt.Axes = None,
        **kwargs):
    """
    Draw a graph in a subplot.

    Parameters
    ----------
    G : nx.DiGraph
        The graph to be drawn.
    layout : dict
        The layout of the graph.
    title : str
        The title of the graph.
    ax : plt.Axes
        The axis in which to draw the graph.
    **formatting_kwargs : dict
        The formatting arguments for the graph.

    Returns
    -------
    None
    """
    edge_colors = list(nx.get_edge_attributes(G, 'color').values())
    widths = list(nx.get_edge_attributes(G, 'width').values())
    styles = list(nx.get_edge_attributes(G, 'style').values())
    default_font_color = 'black'
    # create a dictionary with the default font color of each node
    node_font_color = {node: default_font_color for node in G.nodes}

    def luminance(color):
        r, g, b, _ = color
        return 0.299 * r + 0.587 * g + 0.114 * b

    # Create a colormap list with the colors of the nodes, based on the regr_score
    if all(['regr_score' in G.nodes[node] for node in G.nodes]):
        reg_scores: List[float] = [G.nodes[node]['regr_score']
                                   for node in G.nodes]
        max_cmap_value = max(*reg_scores, 1.0)
        color_map = set_colormap(0.0, max_cmap_value, 'RdYlGn_r')
        kwargs['font_color'] = "black"

        # Set with_labels to False if there is color information of each node,
        # since I will draw the labels afterwards
        kwargs['with_labels'] = False

        # Set the node colors and the label colors according to the value of the
        # regr_score of each node.
        node_colors = []
        linewidths = []
        for node in G:
            node_colors.append(color_map(G.nodes[node]['regr_score']))
            if root_causes is not None and node in root_causes:
                linewidths.append(3.0)
            else:
                linewidths.append(1.0)
            lum = luminance(color_map(G.nodes[node]['regr_score']))
            node_font_color[node] = 'black' if lum > 0.5 else 'white'
        kwargs['node_color'] = node_colors
        kwargs['linewidths'] = linewidths

    nx.draw(G, pos=layout, edge_color=edge_colors,
            width=widths, style=styles, **kwargs, ax=ax)

    if kwargs['with_labels'] is False:
        for _, node in enumerate(G):
            # font_color = label_colors[i]
            nx.draw_networkx_labels(
                G, pos=layout, labels={node: node},
                font_color=node_font_color[node],
                font_weight=kwargs['font_weight'],
                font_family=kwargs['font_family'],
                horizontalalignment=kwargs['horizontalalignment'],
                verticalalignment=kwargs['verticalalignment'],
                ax=ax)

    if title is not None:
        ax.set_title(title, fontsize=12, y=-0.1)


def cleanup_graph(G: nx.DiGraph) -> nx.DiGraph:
    if '\\n' in G.nodes:
        G.remove_node('\\n')
    return G


def set_colormap(
        color_threshold=0.15,
        max_color=0.8,
        cmap_name: str = "OrRd") -> ListedColormap:
    """
    Set the colormap for the graph edges.

    Parameters
    ----------
    color_threshold : float
        The threshold for the color of the values in the plot, below which the color
        will be white.
    max_color : float
        The maximum color for the edges, above which the color will be red.

    Returns
    -------
    LinearColormap
        The colormap to be used in the plot.
    """
    cw = plt.get_cmap(cmap_name)
    cmap = ListedColormap([cw(x)
                           for x in np.arange(color_threshold, max_color, 0.01)])
    cm = copy(cmap)
    cm.set_under(color='white')
    return cm


def dag2dot(
        G: nx.DiGraph,
        undirected=False,
        name: str = "my_dotgraph",
        odots: bool = True) -> Dot:
    """
    Display a DOT of the graph in the notebook.

    Args:
        G (nx.Graph or DiGraph): the graph to be represented.
        undirected (bool): default False, indicates whether the plot is forced
            to contain no arrows.
        plot (bool): default is True, this flag can be used to simply generate
            the object but not plot, in case the object is needed to generate
            a PNG version of the DOT, for instance.
        name (str): the name to be embedded in the Dot object for this graph.
        odots (bool): represent edges with biconnections with circles (odots). if
            this is set to false, then the edge simply has no arrowheads.

    Returns:
        pydot.Dot object
    """
    if len(list(G.edges())) == 0:
        return None
    # Obtain the DOT version of the NX.DiGraph and visualize it.
    if undirected:
        G = G.to_undirected()
        dot_object = nx.nx_pydot.to_pydot(G)
    else:
        # Make a dot Object with edges reflecting biconnections as non-connected edges
        # or arrowheads as circles.
        dot_str = "strict digraph" + name + "{\nconcentrate=true;\n"
        for node in G.nodes():
            dot_str += f"{node};\n"
        if odots:
            options = "[arrowhead=odot, arrowtail=odot, dir=both]"
        else:
            options = "[dir=none]"
        for u, v in G.edges():
            if G.has_edge(v, u):
                dot_str += f"{u} -> {v} {options};\n"
            else:
                dot_str += f"{u} -> {v};\n"
        dot_str += "}\n"
        dot_object = pydotplus.graph_from_dot_data(dot_str)

    # This is to display single arrows with two heads instead of two arrows with
    # one head towards each direction.
    dot_object.set_concentrate(True)
    dot_object.del_node('"\\n"')

    return dot_object


def values_distribution(values, threshold=None, **kwargs):
    """
    Plot the probability density and cumulative density of a given set of values.

    Parameters:
    ----------
        values (array-like): The values to be plotted.
        **kwargs: Additional keyword arguments for customizing the plot.

    Returns:
    -------
        None
    """
    figsize = kwargs.get('figsize', (7, 5))
    _, ax = plt.subplots(1, 2, figsize=figsize)
    ax[0].set_title("Probability density")
    ax[1].set_title("Cumulative density")
    ax[0].set_xlabel("Mean SHAP values")
    ax[1].set_xlabel("Mean SHAP values")
    ax[0].set_ylabel("Probability")
    ax[1].set_ylabel("Cumulative probability")
    sns.histplot(data=values, ax=ax[0], kde=True)
    sns.ecdfplot(data=values, ax=ax[1])
    if threshold is not None:
        ax[1].axvline(threshold, color='red', linestyle='--', linewidth=0.5)
    plt.tight_layout()
    plt.show()


def correlation_matrix(
        corr_matrix: pd.DataFrame,
        sorted_colnames: List[str] = None,
        threshold: float = 0.5,
        ax: axes.Axes = None,
        **kwargs) -> None:
    """
    Plot the correlation matrix of the data.

    Parameters
    ----------
        - corrs (pd.DataFrame)
            Correlation matrix.
        - sorted_colnames (List[str])
            List of sorted column names. If the dataframe contains the names of
                columns already sorted, then no need to pass this argument.
        - threshold (float)
            Threshold for the correlation. Values below this threshold will
                not be displayed
        - ax (matplotlib.axes.Axes)
            Axes to plot the correlation matrix, in case this is a plot to be
                embedded in a subplot. Otherwise, a new figure will be created and
                this argument is not necessary.
        - **kwargs
            Keyword arguments to be passed to the plot_dendogram function.
            - title (str)
                Title of the plot.
            - fontsize (int)
                Font size for the labels.
            - fontname (str)
                Font name for the labels.
            - xrot (int)
                Rotation of the labels.

    Returns
    -------
        None
    """
    if sorted_colnames is None:
        sorted_colnames = corr_matrix.columns

    if ax is None:
        _, ax = plt.subplots()

    title = kwargs.get('title', 'Correlation matrix')
    fontsize = kwargs.get('fontsize', 9)
    fontname = kwargs.get('fontname', "Arial")
    xrot = kwargs.get('xrot', 90)
    cm = set_colormap(color_threshold=threshold, max_color=0.9)
    precision = 2

    def myround(v, ndigits=2):
        if np.isclose(v, 0.0):
            return "0"
        return format(v, '.' + str(ndigits) + 'f')

    corr_data = np.abs(copy(corr_matrix.values))
    ncols, nrows = corr_data.shape
    for x in range(ncols):
        for y in range(nrows):
            if x == y or corr_data[x, y] < threshold:
                corr_data[x, y] = 0

    ax.set_xticks(range(len(sorted_colnames)), sorted_colnames, rotation=xrot,
                  horizontalalignment='center',
                  fontsize=fontsize, fontname=fontname, color='black')
    ax.set_yticks(range(len(sorted_colnames)), sorted_colnames,
                  verticalalignment='center',
                  fontsize=fontsize, fontname=fontname, color='black')
    ax.imshow(corr_data, cmap=cm, vmin=threshold,
              vmax=1.0, aspect='equal')
    ax.grid(True, which='major', alpha=.25)
    for x in range(ncols):
        for y in range(nrows):
            if (x) == y:
                ax.annotate('x', xy=(y, x),
                            horizontalalignment='center',
                            verticalalignment='center',
                            fontsize=fontsize, fontname=fontname, color='black')
            if (x) != y and not np.isclose(round(corr_data[x, y], precision), 0.0):
                ax.annotate(myround(corr_data[x, y], precision), xy=(y, x),
                            horizontalalignment='center',
                            verticalalignment='center',
                            fontsize=fontsize, fontname=fontname, color='black')
    plt.tick_params(pad=10, axis='x', which='both')

    ax.spines['top'].set_linewidth(.3)
    ax.spines['right'].set_linewidth(.3)
    ax.spines['left'].set_linewidth(1)
    ax.spines['left'].set_edgecolor('grey')
    ax.spines['bottom'].set_linewidth(.3)
    ax.set_title(title)


def hierarchies(hierarchies, threshold=0.5, **kwargs):
    """
    Plot the hierarchical clustering and correlation matrix of the data.

    https://www.kaggle.com/code/sgalella/correlation-heatmaps-with-hierarchical-clustering/notebook

    Parameters
    ----------
        - hierarchies (HierarchicalClustering)
            Hierarchical clustering object.
        - threshold (float)
            Threshold for the correlation.
        - **kwargs
            Additional keyword arguments to be passed to the correlation_matrix
                function.

    Returns
    -------
        None
    """
    f_size = kwargs.get('figsize', (9, 4))
    _, (ax1, ax2) = plt.subplots(nrows=1, ncols=2, figsize=f_size)
    dendrogram(hierarchies.linkage_mat, labels=hierarchies.data.columns,
               orientation='top', leaf_rotation=90, ax=ax1)
    ax1.set_title('Hierarchical Clustering Dendrogram')
    ax1.set_ylabel("Dissimilarity")
    ax1.set_ylim(0, 1)

    correlations, sorted_colnames = hierarchies._cluster_features(
        "spearman", threshold)
    correlation_matrix(
        correlations, sorted_colnames, threshold, ax2, **kwargs)


def dag(
        graph: nx.DiGraph,
        reference: nx.DiGraph = None,
        root_causes: list = None,
        show_metrics: bool = False,
        show_node_fill: bool = True,
        title: str = None,
        ax: plt.Axes = None,
        figsize: Tuple[int, int] = (5, 5),
        dpi: int = 75,
        save_to_pdf: str = None,
        layout: str = "dot",
        **kwargs):
    """
    Compare two graphs using dot.

    Parameters:
    -----------
    graph: The DAG to compare.
    reference: The reference DAG.
    root_causes: The root causes of the graph.
    show_metrics: Whether to show the metrics of the graph.
    show_node_fill: Whether to show the node fill (corresponding to the
        root causes).
    title: The title of the graph.
    ax: The axis in which to draw the graph.
    figsize: The size of the figure.
    dpi: The dots per inch of the figure.
    **kwargs: Additional arguments to format the graphs:
        - "node_size": 500
        - "node_color": 'white'
        - "edgecolors": "black"
        - "font_family": "monospace"
        - "horizontalalignment": "center"
        - "verticalalignment": "center_baseline"
        - "with_labels": True
    """
    ncols = 1

    # Overwrite formatting_kwargs with kwargs if they are provided
    formatting_kwargs.update(kwargs)

    # Check consistency
    if show_metrics and reference is None:
        show_metrics = False

    G = nx.DiGraph()
    if show_node_fill:
        G.add_nodes_from(graph.nodes(data=True))
    else:
        G.add_nodes_from(graph.nodes())
    G.add_edges_from(graph.edges())
    if reference:
        # Clean up reference graph for inconsistencies along the DOT conversion
        # and add potential missing nodes to the predicted graph.
        Gt = cleanup_graph(reference.copy())
        for missing in set(list(Gt.nodes)) - set(list(G.nodes)):
            G.add_node(missing)
        G = format_graph(
            G, Gt, inv_color="orange", wrong_color="red", missing_color="lightgrey")
    else:
        G = format_graph(G)

    ref_layout = None
    setup_plot(dpi=dpi)
    if ax is None and show_metrics is False:
        f, axis = plt.subplots(ncols=ncols, figsize=figsize)
    elif ax is None and show_metrics is True:
        metric = evaluate_graph(reference, graph, list(reference.nodes))
        ax = plt.figure(
            figsize=(6, 4), layout="constrained").subplot_mosaic('AAB')
        axis = ax["A"]
        text_axis = ax["B"]
    elif ax is not None and show_metrics is False:
        f = None
        axis = ax
    else:
        raise ValueError(
            "The 'ax' and 'show_metrics' arguments are mutually exclusive.")
    if save_to_pdf is not None:
        with PdfPages(save_to_pdf) as pdf:
            if reference:
                # Added for circular layout compatibility
                Gt = cleanup_graph(reference.copy())
                if layout == "dot":
                    ref_layout = nx.drawing.nx_agraph.graphviz_layout(
                        Gt, prog="dot")
                elif layout == "circular":
                    ref_layout = nx.circular_layout(Gt)  # NEW: Circular layout
                else:
                    raise ValueError(
                        "Invalid layout option. Choose 'dot' or 'circular'.")
            else:
                if layout == "dot":
                    ref_layout = nx.drawing.nx_agraph.graphviz_layout(
                        G, prog="dot")
                elif layout == "circular":
                    ref_layout = nx.circular_layout(G)  # NEW: Circular layout
                else:
                    raise ValueError(
                        "Invalid layout option. Choose 'dot' or 'circular'.")

            draw_graph_subplot(
                G, layout=ref_layout, title=title, ax=axis, **formatting_kwargs)
            pdf.savefig(f, bbox_inches='tight', pad_inches=0)
            plt.close()
    else:
        if reference:
            # added to avoid errors with circular layout
            Gt = cleanup_graph(reference.copy())
            if layout == "dot":
                ref_layout = nx.drawing.nx_agraph.graphviz_layout(
                    Gt, prog="dot")
            elif layout == "circular":
                ref_layout = nx.circular_layout(Gt)  # NEW: Circular layout
            else:
                raise ValueError(
                    "Invalid layout option. Choose 'dot' or 'circular'.")
        else:
            if layout == "dot":
                ref_layout = nx.drawing.nx_agraph.graphviz_layout(
                    G, prog="dot")
            elif layout == "circular":
                ref_layout = nx.circular_layout(G)  # NEW: Circular layout
            else:
                raise ValueError(
                    "Invalid layout option. Choose 'dot' or 'circular'.")

        draw_graph_subplot(
            G, root_causes=root_causes, layout=ref_layout, ax=axis, title=title,
            **formatting_kwargs)

        if show_metrics:
            plt.rcParams["font.family"] = "monospace"
            text_axis.text(0.1, 0.5, metric.matplotlib_repr(),
                           ha='left', va='center', size=12)
            text_axis.axis('off')

        if ax is None:
            plt.show()


def dags(
        dags: List[nx.DiGraph],
        ref_graph: nx.DiGraph,
        titles: List[str],
        figsize: Tuple[int, int] = (15, 12),
        dpi: int = 300) -> None:
    """
    Plots multiple directed acyclic graphs (DAGs) in a grid layout.

    Parameters:
    - dags (list): List of DAGs to plot.
    - ref_graph: Reference graph used for layout.
    - titles (list): List of titles for each DAG.
    - figsize (tuple): Figure size (default: (15, 12)).
    - dpi (int): Dots per inch (default: 300).

    Raises:
    - ValueError: If there are too many DAGs to plot.

    Returns:
    - None
    """
    assert len(titles) == len(
        dags), "Number of titles must match number of DAGs"
    assert len(dags) <= 12 and len(
        dags) > 1, "Number of DAGs must be between 2 and 12"

    layout = _get_subplot_mosaic_layout(len(dags))

    axs = plt.figure(
        figsize=figsize, dpi=dpi, layout="constrained").subplot_mosaic(layout)

    ax_labels = list(axs.keys())
    for i, g in enumerate(dags):
        ax_ = axs[ax_labels[i]]
        dag(graph=g, reference=ref_graph, title=titles[i], ax=ax_)

    plt.tight_layout()
    plt.show()


def _get_subplot_mosaic_layout(n: int) -> str:
    """
    Get a layout string for a subplot mosaic with n subplots.

    The layout strings are hardcoded for 1 to 11 subplots. For more subplots,
    a ValueError is raised.

    Parameters:
    - n (int): The number of subplots.

    Returns:
    - str: The layout string for the subplot mosaic.

    Raises:
    - ValueError: If n is larger than 11.
    """
    layouts = [
        "AB",
        "ABC",
        "AB;CD",
        "AABBCC;.DDEE.",
        "ABC;DEF",
        "ABC;DEF;.G.",
        "ABCD;EFGH",
        "ABC;DEF;GHI",
        "AABBCCDD;EEFFGGHH;.IIJJ.",
        "ABCD;EFGH;IJK.",
        "ABCD;EFGH;IJKL",
    ]
    if n <= len(layouts):
        return layouts[n-2]
    else:
        raise ValueError("Too many DAGs to plot")


def shap_values(shaps: BaseEstimator, **kwargs):
    assert shaps.is_fitted_, "Model not fitted yet"
    plot_args = list(shaps.feature_names)
    return subplots(shaps._plot_shap_summary, *plot_args, **kwargs)


def shap_discrepancies(
        shaps: BaseEstimator,
        target_name: str,
        threshold: float = +100.0,
        regression_line: bool = False,
        reduced: bool = False,
        **kwargs):
    """
    Plot the discrepancies of the SHAP values.

    Parameters:
    -----------
        shaps: BaseEstimator
            The SHAP values to plot.
        target_name: str
            The name of the target variable.
        threshold: float, default=+100.0
            The threshold for the discrepancies. Only discrepancies below this
            threshold will be displayed. Typical values are in (0.0, 5.0), but those
            significant are in the values close to 0.0.
        regression_line: bool, default=False
            If True, include the regression line in the plot.
        Optional arguments:
        - figsize: Tuple[int, int], default=(10, 16)
            The size of the figure.
        - pdf_filename: str, default=None
            The name of the PDF file to save the plot to.
        - dpi: int, default=75
            The DPI of the plot.

    """
    assert shaps.is_fitted_, "Model not fitted yet"
    # shaps._plot_discrepancies(target_name, threshold, **kwargs)

    mpl.rcParams['figure.dpi'] = kwargs.get('dpi', 75)
    pdf_filename = kwargs.get('pdf_filename', None)
    feature_names = [
        f for f in shaps.feature_names if (
            (f != target_name) and
            ((1. - shaps.shap_discrepancies[target_name]
             [f].shap_gof) < threshold)
        )
    ]
    # If no features are found, gracefully return
    if not feature_names:
        print(f"No features with discrepancy index below {threshold} found "
              f"for target {target_name}.")
        return

    # Set the height of the figure to 18 unless there're less than 9 features,
    # in which case, the height is 18/(9 - len(feature_names)).
    if len(feature_names) < 9:
        height = 2*len(feature_names)
    else:
        height = 16
    figsize_ = kwargs.get('figsize', (10, height))

    # The reduced version only plots the scatter and the discrepancies
    nr_axes = 1 if reduced else 4
    fig, ax = plt.subplots(len(feature_names), nr_axes, figsize=figsize_)

    # If the number of features is 1, I must keep ax indexable, so I put it
    # in a list.
    if len(feature_names) == 1:
        ax = [ax]

    for i, parent_name in enumerate(feature_names):
        r = shaps.shap_discrepancies[target_name][parent_name]
        x = shaps.X_test[parent_name].values.reshape(-1, 1)
        y = shaps.X_test[target_name].values.reshape(-1, 1)
        parent_pos = feature_names.index(parent_name)
        s = shaps.shap_scaled_values[target_name][:,
                                                  parent_pos].reshape(-1, 1)
        _plot_discrepancy(
            x, y, s, target_name, parent_name, r, ax[i], regression_line, reduced)

    plt.suptitle(f"Discrepancies for {target_name}")

    if pdf_filename is not None:
        plt.tight_layout(rect=[0, 0.0, 1, 0.97])
        plt.savefig(pdf_filename, bbox_inches='tight')
        plt.close()
    else:
        plt.tight_layout(rect=[0, 0.0, 1, 0.97])
        fig.show()


def _plot_discrepancy(x, y, s, target_name, parent_name, r, ax, regression_line, reduced=False):
    """
    Plot the discrepancy between target and SHAP values.

    Args:
        x (array-like): The x-axis values.
        y (array-like): The target values.
        s (array-like): The SHAP values.
        target_name (str): The name of the target variable.
        parent_name (str): The name of the parent variable.
        r (object): The result object containing model parameters and statistics.
        ax (array-like): The array of subplots.
        regression_line (bool): If True, include the regression line in the plot.
        reduced (bool): If True, plot the reduced version of the SHAP values.

    Returns:
        None
    """
    def _remove_ticks_and_box(ax):
        ax.set_xticks([])
        ax.set_xticklabels([])
        ax.set_yticks([])
        ax.set_yticklabels([])
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        # ax.spines['left'].set_visible(False)

    b0_s, b1_s = r.shap_model.params[0], r.shap_model.params[1]
    b0_y, b1_y = r.parent_model.params[0], r.parent_model.params[1]

    mpl.rc('text', usetex=True)
    mpl.rc('text.latex', preamble=r'\usepackage{amsmath}')

    # If the number of axes is 1, it is not indexable, so I change the only one
    ax_0 = ax if reduced else ax[0]

    # Represent scatter plots
    ax_0.scatter(x, s, alpha=0.5, marker='+')
    ax_0.scatter(x, y, alpha=0.5, marker='.')
    if regression_line:
        ax_0.plot(x, b1_s * x + b0_s, color='blue', linewidth=.5)
        ax_0.plot(x, b1_y * x + b0_y, color='red', linewidth=.5)
    ax_0.set_title(
        fr'$ X_j\, \textrm{{vs.}}\, [X_i | \phi_j] $, '
        fr'\quad $\delta_j^{{(i)}}\!: {1 - r.shap_gof:.2f}$',
        fontsize=10)
    ax_0.set_xlabel(f'${parent_name}$')
    ax_0.set_ylabel(
        fr'$ \mathrm{{{target_name}}} / \phi_{{{parent_name}}} $')

    # Represent target vs. SHAP values
    if not reduced:
        ax[1].scatter(s, y, alpha=0.3, marker='.', color='tab:green')
        ax[1].set_title(
            rf'$\delta_j^{{(i)}}: {1 - r.shap_gof:.2f}$', fontsize=10)
        ax[1].set_xlabel(fr'$ \phi_{{{parent_name}}} $')
        ax[1].set_ylabel(fr'$ \mathrm{{{target_name}}} $')

        # Represent distributions
        pd.DataFrame(s).plot(kind='density', ax=ax[1], label="shap")
        pd.DataFrame(y).plot(kind='density', ax=ax[1], label="parent")
        ax[2].legend().set_visible(False)
        ax[2].set_ylabel('')
        ax[2].set_xlabel(
            fr'$ \mathrm{{{target_name}}} /  \phi_{{{parent_name}}} $')
        ax[2].set_title(rf'$\mathrm{{KS}}({r.ks_pvalue:.2g})$', fontsize=10)

        # Represent fitted vs. residuals
        s_resid = r.shap_model.get_influence().resid_studentized_internal
        y_resid = r.parent_model.get_influence().resid_studentized_internal
        scaler = StandardScaler()
        s_fitted_scaled = scaler.fit_transform(
            r.shap_model.fittedvalues.reshape(-1, 1))
        y_fitted_scaled = scaler.fit_transform(
            r.parent_model.fittedvalues.reshape(-1, 1))
        ax[3].scatter(s_fitted_scaled, s_resid, alpha=0.5, marker='+')
        ax[3].scatter(y_fitted_scaled, y_resid, alpha=0.5,
                      marker='.', color='tab:orange')
        ax[3].set_title(r'$\mathrm{Residuals}$', fontsize=10)
        ax[3].set_xlabel(
            rf'$ \mathrm{{{target_name}}} /  \phi_{{{parent_name}}} $')
        ax[3].set_ylabel(rf'$ \epsilon_{{{target_name}}} / \epsilon_\phi $')

    if not reduced:
        for ax_idx in range(4):
            _remove_ticks_and_box(ax[ax_idx])
    else:
        _remove_ticks_and_box(ax_0)


def deprecated_dags(
        graph: nx.DiGraph,
        reference: nx.DiGraph = None,
        names: List[str] = None,
        figsize: Tuple[int, int] = (10, 5),
        dpi: int = 75,
        save_to_pdf: str = None,
        **kwargs):
    """
    Compare two graphs using dot.

    Parameters:
    -----------
    graph: The DAG to compare.
    reference: The reference DAG.
    names: The names of the reference graph and the dag.
    figsize: The size of the figure.
    **kwargs: Additional arguments to format the graphs:
        - "node_size": 500
        - "node_color": 'white'
        - "edgecolors": "black"
        - "font_family": "monospace"
        - "horizontalalignment": "center"
        - "verticalalignment": "center_baseline"
        - "with_labels": True
    """
    print("This method is deprecated. Use plot_dags() instead.")
    return

    ncols = 1 if reference is None else 2
    if names is None:
        names = ["Prediction", "Ground truth"]

    # Overwrite formatting_kwargs with kwargs if they are provided
    formatting_kwargs.update(kwargs)

    G = nx.DiGraph()
    G.add_nodes_from(graph.nodes(data=True))
    G.add_edges_from(graph.edges())
    if reference:
        # Clean up reference graph for inconsistencies along the DOT conversion
        # and add potential missing nodes to the predicted graph.
        Gt = cleanup_graph(reference.copy())
        for missing in set(list(Gt.nodes)) - set(list(G.nodes)):
            G.add_node(missing)
        # Gt = _format_graph(Gt, Gt, inv_color="red", wrong_color="black")
        # G  = _format_graph(G, Gt, inv_color="red", wrong_color="gray")
        Gt = format_graph(
            Gt, G, inv_color="lightgreen", wrong_color="lightgreen")
        G = format_graph(G, Gt, inv_color="orange", wrong_color="red")
    else:
        G = format_graph(G)

    ref_layout = None
    setup_plot(dpi=dpi)
    f, ax = plt.subplots(ncols=ncols, figsize=figsize)
    ax_graph = ax[1] if reference else ax
    if save_to_pdf is not None:
        with PdfPages(save_to_pdf) as pdf:
            if reference:
                ref_layout = nx.drawing.nx_agraph.graphviz_layout(
                    Gt, prog="dot")
                draw_graph_subplot(Gt, layout=ref_layout, title=None, ax=ax[0],
                                   **formatting_kwargs)
            draw_graph_subplot(G, layout=ref_layout, title=None, ax=ax_graph,
                               **formatting_kwargs)
            pdf.savefig(f, bbox_inches='tight', pad_inches=0)
            plt.close()
    else:
        if reference:
            ref_layout = nx.drawing.nx_agraph.graphviz_layout(
                Gt, prog="dot")
            draw_graph_subplot(Gt, layout=ref_layout, title=names[1], ax=ax[0],
                               **formatting_kwargs)
        draw_graph_subplot(G, layout=ref_layout, title=names[0], ax=ax_graph,
                           **formatting_kwargs)
        plt.show()


#
# Methods below this line were in `plot.py'. I'm moving them here, so they must
# be called now without the leading 'plot_' in the name, but with `plot.` instead.
#

def score_by_method(metrics, metric, methods, **kwargs):
    """
    Plots the score by method.

    Parameters:
    - metrics: DataFrame containing the metrics data.
    - metric: The metric to be plotted.
    - methods: List of methods to be included in the plot.
    - **kwargs: Additional keyword arguments for customization, like
        - figsize: The size of the figure. Default is (4, 3).
        - dpi: The resolution of the figure in dots per inch. Default is 300.
        - title: The title of the plot. Default is None.
        - pdf_filename: The filename to save the plot to. If None, the plot
            will be displayed on screen, otherwise it will be saved to the
        - method_column: The name of the column containing the method names.
            Default is 'method'.

    Returns:
    None
    """
    assert metric in list(score_titles.keys()), \
        ValueError(f"Metric '{metric}' not recognized.")
    figsize_ = kwargs.get('figsize', (4, 3))
    dpi_ = kwargs.get('dpi', 300)
    title_ = kwargs.get('title', None)
    pdf_filename = kwargs.get('pdf_filename', None)
    method_column = kwargs.get('method_column', 'method')

    if methods is None:
        methods = ['rex_mlp', 'rex_gbt', 'rex_intersection', 'rex_union']

    _, ax = plt.subplots(1, 1, figsize=figsize_, dpi=dpi_)

    metric_values = [metrics[metrics[method_column] == m][metric]
                     for m in methods]

    plt.boxplot(metric_values)
    ax.set_xticklabels(labels=[method_labels[key]
                               for key in methods], fontsize=7)
    ax.grid(axis='y', linestyle='--', linewidth=0.5, which='both')

    # check if any value is above 1
    if np.all(np.array(metric_values) <= 1.0) and \
            np.all(np.array(metric_values) >= 0.0):
        ax.set_ylim([0, 1.05])
    if not np.any(np.array(metric_values) < 0.0):
        ax.set_ylim(bottom=0)

    # Set the Y tick labels and left axis bounds
    yticks = ax.get_yticks()
    if np.max(metric_values) <= yticks[-1]:
        yticks = ax.get_yticks()[:-1]
    ax.set_yticklabels(labels=[f"{t:.1f}" for t in yticks], fontsize=6)

    #  Remove the top and right axes
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['left'].set_bounds(high=yticks[-1])
    ax.spines['left'].set_visible(False)

    if title_ is None:
        ax.set_title(score_titles[metric], fontsize=10)
    else:
        ax.set_title(title_, fontsize=10)

    if pdf_filename is not None:
        plt.savefig(pdf_filename, bbox_inches='tight')
        plt.close()
    else:
        plt.tight_layout()
        plt.show()


def scores_by_method(
        metrics: pd.DataFrame,
        methods=None,
        title: str = None,
        pdf_filename=None,
        **kwargs):
    """
    Plot the metrics for all the experiments matching the input pattern

    Parameters
    ----------
    metrics : pandas DataFrame
        A DataFrame with the metrics for all the experiments
    method_types : list
        The list of methods to plot. If None, all the methods will be plotted
        The methods included are: 'rex_mlp', 'rex_gbt', 'rex_intersection' and
        'rex_union'
    title : str
        The title of the plot
    pdf_filename : str
        The filename to save the plot to. If None, the plot will be displayed
        on screen, otherwise it will be saved to the specified filename.

    Optional parameters:
    - figsize (tuple, optional): The size of the figure. Default is (7, 5).
    - dpi (int, optional): The resolution of the figure in dots per inch.
        Default is 300.
    - ylim (tuple, optional): The y-axis limits of the plot. Default is None.
    """
    figsize_ = kwargs.get('figsize', (9, 5))
    dpi_ = kwargs.get('dpi', 300)
    method_column = kwargs.get('method_column', 'method')
    if methods is None:
        methods = ['rex_mlp', 'rex_gbt', 'rex_intersection', 'rex_union']

    what = ['f1', 'precision', 'recall', 'shd', 'sid']
    axs = plt.figure(layout="constrained", figsize=figsize_, dpi=dpi_).\
        subplot_mosaic('AABBCC;.DDEE.')

    ax_labels = list(axs.keys())
    for i, metric in enumerate(what):
        ax = axs[ax_labels[i]]
        metric_values = [metrics[metrics[method_column] == m][metric]
                         for m in methods]

        ax.boxplot(metric_values)
        ax.set_xticklabels(labels=[method_labels[key]
                                   for key in methods], fontsize=6)
        ax.grid(axis='y', linestyle='--', linewidth=0.5, which='both')

        # check if any value is above 1
        if np.all(np.array(metric_values) <= 1.0) and \
                np.all(np.array(metric_values) >= 0.0):
            ax.set_ylim([0, 1.05])
        if not np.any(np.array(metric_values) < 0.0):
            ax.set_ylim(bottom=0)

        # Set the Y tick labels and left axis bounds
        yticks = ax.get_yticks()
        if np.max(metric_values) <= yticks[-1]:
            yticks = ax.get_yticks()[:-1]
        ax.set_yticklabels(labels=[f"{t:.1f}" for t in yticks], fontsize=6)

        #  Remove the top and right axes
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['left'].set_bounds(high=yticks[-1])
        ax.spines['left'].set_visible(False)

        ax.set_title(score_titles[metric], fontsize=10)

    if title is not None:
        fig = plt.gcf()
        fig.suptitle(title)

    if pdf_filename is not None:
        plt.savefig(pdf_filename, bbox_inches='tight')
        plt.close()
    else:
        plt.tight_layout()
        plt.show()


def score_by_subtype(
        metrics: pd.DataFrame,
        score_name: str,
        methods=None,
        pdf_filename=None,
        **kwargs):
    """
    Plots the score by subtype.

    Parameters:
    - metrics (pandas DataFrame): The metrics for all the experiments. This dataframe
    contains the following columns:
        - method (str): The name of the method used.
        - data_type (str): The type of data used.
        - f1 (float): The F1 score.
        - precision (float): The precision score.
        - recall (float): The recall score.
        - aupr (float): The area under the precision-recall curve.
        - Tp (int): The number of true positives.
        - Tn (int): The number of true negatives.
        - Fp (int): The number of false positives.
        - Fn (int): The number of false negatives.
        - shd (int): The structural Hamming distance.
        - sid (int): The structural intervention distance.
        - n_edges (int): The number of edges in the graph.
        - ref_n_edges (int): The number of edges in the reference graph.
        - diff_edges (int): The difference between the number of edges in the graph
        and the reference graph.
        - name (str): The name of the experiment.
    and stores one experiment per row.
    - score_name (str): The name of the score to plot. Valid names are 'f1',
    'precision', 'recall', 'aupr', 'shd', 'sid', 'n_edges', 'ref_n_edges' and
    'diff_edges'.
    - methods (list, optional): The list of methods to plot. If None, all the methods
    will be plotted. The methods included are: 'rex_intersection', 'rex_union',
    'pc', 'fci', 'ges', 'lingam'
    - pdf_filename (str, optional): The filename to save the plot to. If None, the plot
    will be displayed on screen, otherwise it will be saved to the specified filename.

    Optional parameters:
    - figsize (tuple, optional): The size of the figure. Default is (2, 1).
    - dpi (int, optional): The resolution of the figure in dots per inch. Default is 300.
    - method_column (str, optional): The name of the column in the metrics dataframe
        that contains the method name. Default is 'method'.

    Returns:
    None
    """
    figsize_ = kwargs.get('figsize', (9, 5))
    dpi_ = kwargs.get('dpi', 300)
    method_column = kwargs.get('method_column', 'method')

    if methods is None:
        methods = ['pc', 'fci', 'ges', 'lingam', 'cam', 'notears', 'un_G_iter']
    x_labels = [method_labels[m] for m in methods]
    axs = plt.figure(layout="constrained", figsize=figsize_, dpi=dpi_).\
        subplot_mosaic('AABBCC;.EEFF.')

    # Loop through all the subtypes
    ax_labels = list(axs.keys())
    for i, subtype in enumerate(synth_data_types):  # + ['all']):
        ax = axs[ax_labels[i]]
        if subtype == 'all':
            sub_df = metrics
        else:
            sub_df = metrics[metrics['data_type'] == subtype]
        metric_values = [sub_df[sub_df[method_column] == m][score_name]
                         for m in methods]
        ax.boxplot(metric_values)
        ax.set_xticklabels(labels=x_labels, fontsize=6)

        if np.all(np.array(metric_values) <= 1.0) and \
                np.all(np.array(metric_values) >= 0.0):
            ax.set_ylim([0, 1.05])
        if not np.any(np.array(metric_values) < 0.0):
            ax.set_ylim(bottom=0)

        yticks = ax.get_yticks()
        if np.max(metric_values) <= yticks[-2]:
            yticks = ax.get_yticks()[:-1]
        ax.set_yticklabels(labels=[f"{t:.1f}" for t in yticks], fontsize=6)

        ax.grid(axis='y', linestyle='--', linewidth=0.5, which='both')

        #  Remove the top and right axes
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['left'].set_bounds(high=yticks[-1])
        ax.spines['left'].set_visible(False)

        # Set the title to the score name, in Latex math mode
        if subtype == 'all':
            ax.set_title(
                r'$\textrm{{all data}}$',
                fontsize=10)  # , y=-0.25)
        else:
            ax.set_title(
                rf'{synth_data_labels[synth_data_types.index(subtype)]} data',
                fontsize=10)  # , y=-0.25)

    plt.suptitle(f"{score_titles[score_name]} score")
    if pdf_filename is not None:
        plt.savefig(pdf_filename, bbox_inches='tight')
        plt.close()
    else:
        plt.tight_layout()
        plt.show()


def combined_metrics(
        metrics: dict,
        metric_types: list = None,
        title: str = None,
        acyclic=False,
        medians=False,
        pdf_filename=None):
    """
    Plot the metrics for all the experiments matching the input pattern

    Parameters
    ----------
    metrics : dict
        A dictionary with the metrics for all the experiments
    title : str
        The title of the plot
    acyclic : bool
        Whether to plot the metrics for the no_cycles graphs
    medians: bool
        Whether to plot the median lines

    Returns
    -------
    None
    """
    what = ['f1', 'precision', 'recall', 'shd', 'sid']
    axs = plt.figure(layout="constrained").subplot_mosaic(
        """
        AABBCC
        .DDEE.
        """
    )
    # fig, axs = plt.subplots(2, 2, figsize=(6, 5))
    if metric_types is None:
        if acyclic:
            metric_types = global_nc_metric_types
        else:
            metric_types = global_metric_types

    ax_labels = list(axs.keys())
    for i, metric in enumerate(what):
        # row, col = i // 2, i % 2
        ax = axs[ax_labels[i]]
        metric_values = [metrics[key].loc[:, metric] for key in metric_types]

        if medians:
            combined_median = np.median(metrics['intersection'].loc[:, metric])
            added_median = np.median(metrics['union'].loc[:, metric])
            ax.axhline(
                combined_median, color='g', linestyle='--', linewidth=0.5)
            ax.axhline(
                added_median, color='b', linestyle='--', linewidth=0.5)

        ax.boxplot(
            metric_values,
            labels=[metric_labels[key] for key in metric_types])
        # check if any value is above 1
        if np.all(np.array(metric_values) <= 1.0):
            ax.set_ylim([0, 1.05])
        ax.set_title(score_titles[metric])

    if title is not None:
        fig = plt.gcf()
        fig.suptitle(title)

    if pdf_filename is not None:
        plt.savefig(pdf_filename, bbox_inches='tight')
        plt.close()
    else:
        plt.tight_layout()
        plt.show()


def _format_mean_std(data):
    # """\scalemath{0.6}{\ \pm\ 0.05}"""
    return rf'${data.mean():.2f} \scalemath{{0.6}}{{\ \pm\ {data.std():.1f}}}$'


def latex_table_by_datatype(df, method, metrics=None):
    if metrics is None:
        metrics = ['precision', 'recall', 'f1', 'shd', 'sid']

    table = "\\begin{tabular}{l|" + 'c'*len(metrics) + "}\n\\toprule\n"
    # table += "{} & Precision & Recall & F1 & SHD & SID \\\\ \\midrule\n"
    table += "{} " + \
        ''.join(
            f"& {score_titles[m]}" for m in metrics) + " \\\\ \\midrule\n"
    for i, data_type in enumerate(synth_data_types):
        table += synth_data_labels[i]
        for metric in metrics:
            data = df[(df.method == method) & (
                df.data_type == data_type)][metric]
            table += f" & {_format_mean_std(data)}"
        table += " \\\\\n"
    table += "\\bottomrule\n"
    table += "\\end{tabular}"
    print(table)


def latex_table_by_method(df, methods=None, metric_names=None):

    if methods is None:
        methods = ['rex_mlp', 'rex_gbt', 'rex_union',
                   'rex_union_adjnc', 'pc', 'fci', 'ges', 'lingam']

    if metric_names is None:
        metric_names = ['precision', 'recall', 'f1', 'shd', 'sid']

    table = "\\begin{tabular}{l|" + 'c'*len(metric_names) + "}\n\\toprule\n"
    table += "{} " + \
        ''.join(
            f"& {score_titles[m]}" for m in metric_names) + " \\\\ \\midrule\n"
    for method in methods:
        table += method_labels[method]
        for metric in metric_names:
            data = df[(df.method == method)][metric]
            table += f" & {_format_mean_std(data)}"
        table += " \\\\\n"
    table += "\\bottomrule\n"
    table += "\\end{tabular}"
    print(table)
